/*  weblog SQL HTTPD log parser
    Copyright (C) 2002 Mathijs de Bruin
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
	    
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
						
#include "weblog.h"

/*
    Although this should actually be moved to weblog.h,
    I thought it would be more handy to leave them here :P. 
    TODO: make ALL the settings go nicely into that one struct
*/
char* dbname;
char* dbhost;
char* dbuser;
char* dbpass;
char* dbsock;
char* conf;
int   dbport, no_stamps;

struct {
    unsigned dbname:1, dbhost:1, dbuser:1, dbpass:1, dbsock:1, no_stamps:1, conf:1, dbport:1;
} set;

MYSQL sqlconn;

void die(char *err) {
    fprintf(stderr,"\nError: %s\nQuitting...\n", err);
    exit(EXIT_FAILURE);
}
    
void sql_connect() {
    mysql_init(&sqlconn);

    printf("Connecting with database %s on %s...", dbname, dbhost);
    fflush(stdout);
    if (!mysql_real_connect(&sqlconn, dbhost, dbuser, dbpass, dbname, dbport, dbsock, 0))
	die(mysql_error(&sqlconn));
	
    printf(" done\n");
}
    
void sql_query(char *query) {
    unsigned long numrows;
    if (mysql_query(&sqlconn, query))
	die(mysql_error(&sqlconn));
    
    numrows = (unsigned long)mysql_affected_rows(&sqlconn);
    if (numrows == 1)
	printf(" 1 row affected\n");
    else
	printf(" %lu rows affected\n",numrows);
}
    
void parse_opts(int argc, char **argv) {
    int opt;

    set.dbhost = 0;
    set.dbuser = 0;
    set.dbpass = 0;
    set.dbname = 0;
    set.dbsock = 0;
    set.dbport = 0;
    set.no_stamps = 0;
    
    while ((opt = getopt(argc, argv, "ic:h:u:p:d:s:P:")) != -1) {
	switch (opt) {
    	    case ':': case '?':
		fprintf(stderr,"usage: weblog [-i] [-c file] [-h host] [-u user] [-p password] [-d database] [-s socket]\n\n");
		fprintf(stderr,"   -i           resolve all hostnames (ignore timestamps)\n");
		fprintf(stderr,"   -c file	configuration file\n");
		fprintf(stderr,"   -h host      database host\n");
		fprintf(stderr,"   -u user      datanase user\n");
		fprintf(stderr,"   -p password  database password\n");
		fprintf(stderr,"   -d database  database name\n");
		fprintf(stderr,"   -s socket    database socket path\n");
		fprintf(stderr,"   -P port      port number to connect to\n");
		exit(EXIT_FAILURE);
	    case 'i':
		no_stamps = 1;
		break;
	    case 'c':
		conf = optarg;
		set.conf = 1;
		break;
	    case 'h':
		dbhost = optarg;
		set.dbhost = 1;
		break;
	    case 'u':
		dbuser = optarg;
		set.dbuser = 1;
		break;
	    case 'p':
		dbpass = optarg;
		set.dbpass = 1;
		break;
	    case 'd':
		dbname = optarg;
		set.dbname = 1;
		break;
	    case 's':
		dbsock = optarg;
		set.dbsock = 1;
		break;
	    case 'P':
		dbport = (unsigned int)strtol(optarg, NULL, 0);
		set.dbport = 1;
	}
    }
}

void parse_conf() {
    CONFIG *cnf;
    char *val;
    int result;
    
    if (!set.conf)
	conf = CONF;	
    
    printf("Reading configuration file %s... ", conf);
    fflush(stdout);

    cnf = init_config();
    result = get_config(cnf, conf, " ");

    if (result == 1) {
	printf(" failed\n");
	fprintf(stderr,"Warning: Configuration file not found, continuing with program defaults.\n");
    }

    if (result == 0) printf(" done\n");
    
    if (result == 2) die("Syntax error reading configuration file");
        
    val = print_config(cnf, "DBHost");
    // Now THIS is what I call UGLY code ;)
    if (!set.dbhost && (dbhost = DBHOST) && result != 1 && val != NULL)
	dbhost = val;
    
    val = print_config(cnf, "DBUser");
    if (!set.dbuser && (dbuser = DBUSER) && result != 1 && val != NULL)
	dbuser = val;

    val = print_config(cnf, "DBPass");
    if (!set.dbpass && (dbpass = DBPASS) && result != 1 && val != NULL) {
    	struct stat cnfstat;

	if (!stat(conf, &cnfstat) && !(cnfstat.st_mode & S_IRWXO))
	    dbpass = val;
	else
	    fprintf(stderr,"Warning: Configuration file world readable, using password from defaults.\n");
	    dbpass = DBPASS;
    }
    	
    val = print_config(cnf, "DBName");
    if (!set.dbname && (dbname = DBNAME) && result != 1 && val != NULL)
	dbname = val;
	
    val = print_config(cnf, "DBSock");
    if (!set.dbsock && (dbsock = DBSOCK) && result != 1 && val != NULL)
	dbsock = val;

    val = print_config(cnf, "DBPort");
    if (!set.dbport && (dbhost = DBHOST) && result != 1 && val != NULL)
	dbport = (unsigned int)strtol(val, NULL, 0);

    val = print_config(cnf, "NoTimestamps");
    if (!set.no_stamps && (no_stamps = 0) && result != 1 && val != NULL && val != "0")
	no_stamps = 1;

    printf("\n");
    
    // This is the cause of some nasty segfaults ;)
    //destroy_config(cnf);
}

int main(int argc, char **argv) {
    unsigned long date, count, total;
    struct hostent *host;
    unsigned int addr;
    char *query;
    
    MYSQL_RES *result;
    MYSQL_ROW row;

    query = malloc(255);
    
    printf("%s version %s\n", PACKAGE, VERSION);
    printf("Copyright (C) 2002 Mathijs de Bruin\n\n");
    
    // Pase command line arguments    
    parse_opts(argc, argv);

    // Parse configuration file / set defaults
    parse_conf();
        
    // Connect to the database
    sql_connect();
    
    // Acquiring current date for later use
    printf("\nFetching datestamp for later use...");
    fflush(stdout);

    if (mysql_query(&sqlconn, "SELECT CURRENT_DATE()+0;"))
	die(mysql_error(&sqlconn));
	
    if ((result = mysql_store_result(&sqlconn))==NULL)
	die(mysql_error(&sqlconn));
	
    row = mysql_fetch_row(result);
    date = atoi(row[0]);
    printf(" done\n");


    // Update the data
    printf("\nUpdating references:\n");
    printf("Agents...");
    fflush(stdout);
    sql_query("INSERT INTO agent (agent_string) SELECT DISTINCT agent FROM rawlog;");

    printf("Virtual hosts...");
    fflush(stdout);
    sql_query("INSERT INTO vhost (vhost_string) SELECT DISTINCT virtual_host FROM rawlog WHERE status=200;");

    printf("Remote hosts...");
    fflush(stdout);
    sql_query("INSERT INTO host (ip) SELECT DISTINCT INET_ATON(remote_host) FROM rawlog;");

    printf("Referers...");
    fflush(stdout);
    sql_query("INSERT INTO referer (referer_string) SELECT DISTINCT referer FROM rawlog;");

    printf("URI's...");
    fflush(stdout);
    sql_query("INSERT INTO uri (uri_string) SELECT DISTINCT request_uri FROM rawlog;");

    printf("HTTP statusses...");
    fflush(stdout);
    sql_query("INSERT INTO status (status_id) SELECT DISTINCT status FROM rawlog;");
    

    printf("\nGenerating further statistics:\n");
    printf("Agents...");
    fflush(stdout);
    sql_query("REPLACE INTO stats SELECT FROM_UNIXTIME(rawlog.time_stamp,'%Y-%m-%d') AS date, vhost.vhost_id, agent.agent_id AS data_id, 'agent' AS data_type, COUNT(*) AS hits FROM rawlog, vhost, agent WHERE vhost.vhost_string = rawlog.virtual_host AND agent.agent_string = rawlog.agent GROUP BY date, vhost_id, data_id;");

    printf("Remote hosts...");
    fflush(stdout);
    sql_query("REPLACE INTO stats SELECT FROM_UNIXTIME(rawlog.time_stamp,'%Y-%m-%d') AS date, vhost.vhost_id, host.host_id AS data_id, 'host' AS data_type, COUNT(*) AS hits FROM rawlog, vhost, host WHERE vhost.vhost_string = rawlog.virtual_host AND host.ip = INET_ATON(rawlog.remote_host) GROUP BY date, vhost_id, data_id;");

    printf("Referers...");
    fflush(stdout);
    sql_query("REPLACE INTO stats SELECT FROM_UNIXTIME(rawlog.time_stamp,'%Y-%m-%d') AS date, vhost.vhost_id, referer.referer_id AS data_id, 'referer' AS data_type, COUNT(*) AS hits FROM rawlog, vhost, referer WHERE vhost.vhost_string = rawlog.virtual_host AND referer.referer_string = rawlog.referer GROUP BY date, vhost_id, data_id;");

    printf("URI's...");
    fflush(stdout);
    sql_query("REPLACE INTO stats SELECT FROM_UNIXTIME(rawlog.time_stamp,'%Y-%m-%d') AS date, vhost.vhost_id, uri.uri_id AS data_id, 'uri' AS data_type, COUNT(*) AS hits FROM rawlog, vhost, uri WHERE vhost.vhost_string = rawlog.virtual_host AND uri.uri_string = rawlog.request_uri GROUP BY date, vhost_id, data_id;");

    printf("HTTP statusses...");
    fflush(stdout);
    sql_query("REPLACE INTO stats SELECT FROM_UNIXTIME(rawlog.time_stamp,'%Y-%m-%d') AS date, vhost.vhost_id, status.status_id AS data_id, 'status' AS data_type, COUNT(*) AS hits FROM rawlog, vhost, status WHERE vhost.vhost_string = rawlog.virtual_host AND status.status_id = rawlog.status GROUP BY date, vhost_id, data_id;");
    
    printf("\nGenerating general traffic statistics...");
    fflush(stdout);
    sql_query("REPLACE INTO visits SELECT FROM_UNIXTIME(rawlog.time_stamp,'%Y-%m-%d') AS date, vhost.vhost_id, FROM_UNIXTIME(rawlog.time_stamp,'%H') AS hour, COUNT(*) AS hits, COUNT(DISTINCT rawlog.remote_host) AS visits, SUM(rawlog.bytes_sent) AS traffic FROM rawlog, vhost WHERE vhost.vhost_string = rawlog.virtual_host GROUP BY date,hour,vhost_id;");
    

    // Resolve hostnames
    printf("\nResolving hostnames:\n");
    printf("Retrieving list of hosts to resolve...");
    fflush(stdout);

    // If option -i is used, hostname timestamps are ignored
    if (no_stamps) {
	snprintf(query, 255, "SELECT ip FROM host WHERE hostname IS NULL");
    } else {
	snprintf(query, 255, "SELECT ip FROM host WHERE hostname IS NULL AND time_stamp > NOW() - INTERVAL %d DAY;", EXPIRE);
    }

    if (mysql_query(&sqlconn, query))
	die(mysql_error(&sqlconn));

    result = mysql_store_result(&sqlconn);
    total = (unsigned long)mysql_num_rows(result);
    printf(" %lu hosts retrieved\n", total);
    
    printf("Resolving hostnames... [  0%]");
    fflush(stdout);
    sethostent(1);
    count = 0;
    
    while (row = mysql_fetch_row(result)) {
	addr = inet_addr(row[0]);
	host = gethostbyaddr((char *)&addr, sizeof(addr), AF_INET);
	count++;
	printf("\b\b\b\b\b\b[%3d%%]",(unsigned int)(((float)count/total)*100));
	fflush(stdout);
	if (host) {
	    snprintf(query, 255, "UPDATE host SET hostname='%s' WHERE ip=%s;", host->h_name, row[0]);
	    if (mysql_query(&sqlconn, query))
		die(mysql_error(&sqlconn));
	} else {
	    // Insert the raw IP number if there's no use in trying again.
	    if (h_errno != TRY_AGAIN) {
		snprintf(query, 255, "UPDATE host SET hostname='%s' WHERE ip=%s;", row[0], row[0]);
		if (mysql_query(&sqlconn, query))
		    die(mysql_error(&sqlconn));
	    // I'm removing this ugly error message cause it isn't of any use anyway...
	    //} else {
	    //	fprintf(stderr, "\nError: %s", hstrerror(h_errno));
	    }
	}
    }
    printf("\b\b\b\b\b\bdone  \n");
    endhostent();
    mysql_free_result(result);
    
    // Update host_info table where needed
    printf("Updating top-level domains...");
    fflush(stdout);

    sql_query("INSERT INTO host_info (host_suffix) SELECT DISTINCT LCASE(RIGHT(hostname, LOCATE('.',REVERSE(hostname)))) AS host_suffix FROM host WHERE hostname IS NOT NULL;");

    // Purging old logs to reduce parse time
    printf("\nPuriging old raw log:\n");   
     
    // Backup the to-be-deleted records
    printf("Backup up old records...");
    fflush(stdout);
    snprintf(query, 255, "INSERT INTO backup SELECT * FROM rawlog WHERE FROM_UNIXTIME(time_stamp, '%%Y%%m%%d') <= %lu;", date-1);
    sql_query(query);
    
    // Delete old records
    printf("Deleting old records...");
    fflush(stdout);

    if ((unsigned long)mysql_affected_rows(&sqlconn)) {
	snprintf(query, 255, "DELETE FROM rawlog WHERE FROM_UNIXTIME(time_stamp, '%%Y%%m%%d') <= %lu;", date-1);
	sql_query(query);

	// Optimize the rawlog table for faster processing next time
	printf("Optimising rawlog table...");
	fflush(stdout);
	if (mysql_query(&sqlconn, "OPTIMIZE TABLE rawlog;"))
	    die(mysql_error(&sqlconn));
	printf(" done\n");
    } else {
	printf(" not needed\n");
    }
        
    // Close database and exit
    printf("\nClosing database connection.\n");
    mysql_close(&sqlconn);
							     
    return EXIT_SUCCESS;
}
