#
# Create database 'weblog' if not already existent
#

CREATE DATABASE /*!32312 IF NOT EXISTS*/ weblog;

USE weblog;

#
# Table structure for table 'rawlog' and backup table 'backup'
#

CREATE TABLE rawlog (
    time_stamp int(10) unsigned default NULL,
    virtual_host char(50) NOT NULL default '0',
    request_uri char(255) default NULL,
    referer char(255) default NULL,
    remote_host char(15) NOT NULL default '',
    bytes_sent int(11) unsigned NOT NULL default '0',
    status smallint(3) unsigned NOT NULL default '0',
    agent char(255) default NULL,
    KEY (time_stamp,virtual_host)
) TYPE=MyISAM;
		  
CREATE TABLE backup (
    time_stamp int(10) unsigned default NULL,
    virtual_host char(50) NOT NULL default '0',
    request_uri char(255) default NULL,
    referer char(255) default NULL,
    remote_host char(15) NOT NULL default '',
    bytes_sent int(11) unsigned NOT NULL default '0',
    status smallint(3) unsigned NOT NULL default '0',
    agent char(255) default NULL
) TYPE=MyISAM;		  

#
# Table structure for table 'agent'
#

CREATE TABLE agent (
    agent_id smallint(5) unsigned NOT NULL auto_increment,
    agent_string varchar(255) default NULL,
    os varchar(128) default NULL,
    browser varchar(128) default NULL,
    PRIMARY KEY  (agent_id),
    UNIQUE KEY agent_string (agent_string)
) TYPE=MyISAM COMMENT='List of agents';
	    
	    
#
# Table structure for table 'host'
#

CREATE TABLE host (
    host_id smallint(5) unsigned NOT NULL auto_increment,
    hostname varchar(255) default NULL,
    ip int(10) unsigned NOT NULL default '0',
    PRIMARY KEY  (host_id,ip),
    UNIQUE KEY host_string (ip),
    time_stamp TIMESTAMP
) TYPE=MyISAM COMMENT='Remote hosts';


#
# Table structure for table 'host_info'
#

CREATE TABLE host_info (
    host_suffix varchar(4) NOT NULL default '',
    info varchar(128) default NULL,
    PRIMARY KEY  (host_suffix)
) TYPE=MyISAM COMMENT='Info per host-suffix (countries etc.)';


#
# Table structure for table 'referer'
#

CREATE TABLE referer (
    referer_id smallint(5) unsigned NOT NULL auto_increment,
    referer_string varchar(255) default NULL,
    PRIMARY KEY  (referer_id),
    UNIQUE KEY referer_string (referer_string)
) TYPE=MyISAM COMMENT='Request''s referers';


#
# Table structure for table 'stats'
#

CREATE TABLE stats (
    date date NOT NULL default '0000-00-00',
    vhost_id smallint(5) unsigned NOT NULL default '0',
    data_id smallint(5) unsigned NOT NULL default '0',
    data_type enum('agent','uri','status','referer','host') NOT NULL default 'agent',
    hits smallint(5) unsigned default '0',
    PRIMARY KEY  (data_id,data_type,date,vhost_id)
) TYPE=MyISAM COMMENT='General statistics';


#
# Table structure for table 'status'
#

CREATE TABLE status (
    status_id smallint(5) unsigned NOT NULL default '0',
    status varchar(50) default NULL,
    PRIMARY KEY  (status_id),
    UNIQUE KEY status (status)
) TYPE=MyISAM COMMENT='HTTP Status codes';


#
# Table structure for table 'uri'
#

CREATE TABLE uri (
    uri_id smallint(5) unsigned NOT NULL auto_increment,
    uri_string varchar(255) default NULL,
    PRIMARY KEY  (uri_id),
    UNIQUE KEY uri_string (uri_string)
) TYPE=MyISAM COMMENT='Requested uri''s';

#
# Table structure for table 'vhost'
#

CREATE TABLE vhost (
    vhost_id smallint(5) unsigned NOT NULL auto_increment,
    vhost_string varchar(50) default NULL,
    PRIMARY KEY  (vhost_id),
    UNIQUE KEY domain_string (vhost_string)
) TYPE=MyISAM COMMENT='Vhosts';


#
# Table structure for table 'visits'
#

CREATE TABLE visits (
    date date NOT NULL default '0000-00-00',
    vhost_id smallint(5) unsigned NOT NULL default '0',
    hour tinyint(3) unsigned NOT NULL default '0',
    hits smallint(5) unsigned default NULL,
    visits smallint(5) unsigned default NULL,
    traffic int(10) unsigned default NULL,
    PRIMARY KEY  (date,vhost_id,hour)
) TYPE=MyISAM COMMENT='Stats per hour';

#
# Adding required users with the right permissions.
# Passwords will be set later by a separate script.
#

#
# Add user parser which is used to parse the 'rawlog' table
# into the relational tables of the weblog table.
#

GRANT USAGE ON *.* TO weblog@localhost;

#
# The user weblog is allowed to perform almost all actions on the weblog database.
#

GRANT SELECT, INSERT, UPDATE, DELETE ON weblog.* TO 'weblog';


#
# The user www will be the one logging to the 'rawlog' table 
# and reading the parsed logs from the rest of the weblog database.
#

GRANT USAGE ON *.* TO www@localhost;

#
# User www will be able to INSERT into rawlog and to SELECT from all the tables
#

GRANT INSERT ON weblog.rawlog TO 'www';
GRANT SELECT ON weblog.* TO 'www';


#
# Apply the privileges
#

FLUSH PRIVILEGES;


#
# Filling host_info table
#

INSERT INTO host_info VALUES ('.ac','Ascension Island'),('.ad','Andorra'),('.ae','United Arab Emirates'),('.af','Afghanistan'),('.ag','Antigua and Barbuda'),('.ai','Anguilla'),('.al','Albania'),('.am','Armenia'),('.an','Netherlands Antilles'),('.ao','Angola'),('.aq','Antarctica'),('.ar','Argentina'),('.as','American Samoa'),('.at','Austria'),('.au','Australia'),('.aw','Aruba'),('.az','Azerbaijan'),('.ba','Bosnia and Herzegovina'),('.bb','Barbados'),('.bd','Bangladesh'),('.be','Belgium'),('.bf','Burkina Faso'),('.bg','Bulgaria'),('.bh','Bahrain'),('.bi','Burundi'),('.bj','Benin'),('.bm','Bermuda'),('.bn','Brunei Darussalam'),('.bo','Bolivia'),('.br','Brazil'),('.bs','Bahamas'),('.bt','Bhutan'),('.bv','Bouvet Island'),('.bw','Botswana'),('.by','Belarus'),('.bz','Belize'),('.ca','Canada'),('.cc','Cocos (Keeling) Islands'),('.cd','Congo'),('.cf','Central African Republic'),('.cg','Congo'),('.ch','Switzerland'),('.ci','Cote d\'Ivoire'),('.ck','Cook Islands'),('.cl','Chile'),('.cm','Cameroon'),('.cn','China'),('.co','Colombia'),('.cr','Costa Rica'),('.cu','Cuba'),('.cv','Cap Verde'),('.cx','Christmas Island'),('.cy','Cyprus'),('.cz','Czech Republic'),('.de','Germany'),('.dj','Djibouti'),('.dk','Denmark'),('.dm','Dominica'),('.do','Dominican Republic'),('.dz','Algeria'),('.ec','Ecuador'),('.ee','Estonia'),('.eg','Egypt'),('.eh','Western Sahara'),('.er','Eritrea'),('.es','Spain'),('.et','Ethiopia'),('.fi','Finland'),('.fj','Fiji'),('.fk','Falkland Islands (Malvina)'),('.fm','Micronesia'),('.fo','Faroe Islands'),('.fr','France'),('.ga','Gabon'),('.gd','Grenada'),('.ge','Georgia'),('.gf','French Guiana'),('.gg','Guernsey'),('.gh','Ghana'),('.gi','Gibraltar'),('.gl','Greenland'),('.gm','Gambia'),('.gn','Guinea'),('.gp','Guadeloupe'),('.gq','Equatorial Guinea'),('.gr','Greece'),('.gs','South Georgia and the South Sandwich Islands'),('.gt','Guatemala'),('.gu','Guam'),('.gw','Guinea-Bissau'),('.gy','Guyana'),('.hk','Hong Kong'),('.hm','Heard and McDonald Islands'),('.hn','Honduras'),('.hr','Croatia/Hrvatska'),('.ht','Haiti'),('.hu','Hungary'),('.id','Indonesia'),('.ie','Ireland'),('.il','Israel'),('.im','Isle of Man'),('.in','India'),('.io','British Indian Ocean Territory'),('.iq','Iraq'),('.ir','Iran (Islamic Republic of)'),('.is','Iceland'),('.it','Italy'),('.je','Jersey'),('.jm','Jamaica'),('.jo','Jordan'),('.jp','Japan'),('.ke','Kenya'),('.kg','Kyrgyzstan'),('.kh','Cambodia'),('.ki','Kiribati'),('.km','Comoros'),('.kn','Saint Kitts and Nevis'),('.kp','Korea'),('.kr','Korea'),('.kw','Kuwait'),('.ky','Cayman Islands'),('.kz','Kazakhstan'),('.la','Lao People\'s Democratic Republic'),('.lb','Lebanon'),('.lc','Saint Lucia'),('.li','Liechtenstein'),('.lk','Sri Lanka'),('.lr','Liberia'),('.ls','Lesotho'),('.lt','Lithuania'),('.lu','Luxembourg'),('.lv','Latvia'),('.ly','Libyan Arab Jamahiriya'),('.ma','Morocco'),('.mc','Monaco'),('.md','Moldova'),('.mg','Madagascar'),('.mh','Marshall Islands'),('.mk','Macedonia'),('.ml','Mali'),('.mm','Myanmar'),('.mn','Mongolia'),('.mo','Macau'),('.mp','Northern Mariana Islands'),('.mq','Martinique'),('.mr','Mauritania'),('.ms','Montserrat'),('.mt','Malta'),('.mu','Mauritius'),('.mv','Maldives'),('.mw','Malawi'),('.mx','Mexico'),('.my','Malaysia'),('.mz','Mozambique'),('.na','Namibia'),('.nc','New Caledonia'),('.ne','Niger'),('.nf','Norfolk Island'),('.ng','Nigeria'),('.ni','Nicaragua'),('.nl','Netherlands'),('.no','Norway'),('.np','Nepal'),('.nr','Nauru'),('.nu','Niue'),('.nz','New Zealand'),('.om','Oman'),('.pa','Panama'),('.pe','Peru'),('.pf','French Polynesia'),('.pg','Papua New Guinea'),('.ph','Philippines'),('.pk','Pakistan'),('.pl','Poland'),('.pm','St. Pierre and Miquelon'),('.pn','Pitcairn Island'),('.pr','Puerto Rico'),('.ps','Palestinian Territories'),('.pt','Portugal'),('.pw','Palau'),('.py','Paraguay'),('.qa','Qatar'),('.re','Reunion Island'),('.ro','Romania'),('.ru','Russian Federation'),('.rw','Rwanda'),('.sa','Saudi Arabia'),('.sb','Solomon Islands'),('.sc','Seychelles'),('.sd','Sudan'),('.se','Sweden'),('.sg','Singapore'),('.sh','St. Helena'),('.si','Slovenia'),('.sj','Svalbard and Jan Mayen Islands'),('.sk','Slovak Republic'),('.sl','Sierra Leone'),('.sm','San Marino'),('.sn','Senegal'),('.so','Somalia'),('.sr','Suriname'),('.st','Sao Tome and Principe'),('.sv','El Salvador'),('.sy','Syrian Arab Republic'),('.sz','Swaziland'),('.tc','Turks and Caicos Islands'),('.td','Chad'),('.tf','French Southern Territories'),('.tg','Togo'),('.th','Thailand'),('.tj','Tajikistan'),('.tk','Tokelau'),('.tm','Turkmenistan'),('.tn','Tunisia'),('.to','Tonga'),('.tp','East Timor'),('.tr','Turkey'),('.tt','Trinidad and Tobago'),('.tv','Tuvalu'),('.tw','Taiwan'),('.tz','Tanzania'),('.ua','Ukraine'),('.ug','Uganda'),('.uk','United Kingdom'),('.um','US Minor Outlying Islands'),('.us','United States'),('.uy','Uruguay'),('.uz','Uzbekistan'),('.va','Holy See (City Vatican State)'),('.vc','Saint Vincent and the Grenadines'),('.ve','Venezuela'),('.vg','Virgin Islands (British)'),('.vi','Virgin Islands (USA)'),('.vn','Vietnam'),('.vu','Vanuatu'),('.wf','Wallis and Futuna Islands'),('.ws','Western Samoa'),('.ye','Yemen'),('.yt','Mayotte'),('.yu','Yugoslavia'),('.za','South Africa'),('.zm','Zambia'),('.zw','Zimbabwe'),('.biz','Businesses'),('.com','Commercial'),('.edu','Educational'),('.gov','US Government'),('.inf','International Organizations'),('.mil','US Dept of Defense'),('.net','Networks'),('.org','Organisations');

#
# Filling status table
#

INSERT INTO status VALUES (100,'Continue'),(101,'Switching Protocols'),(200,'OK'),(201,'Created'),(202,'Accepted'),(203,'Non-Authoritative Information'),(204,'No Content'),(205,'Reset Content'),(206,'Partial Content'),(300,'Multiple Choices'),(301,'Moved Permanently'),(302,'Found'),(303,'See Other'),(304,'Not Modified'),(305,'Use Proxy'),(307,'Temporary Redirect'),(400,'Bad Request'),(401,'Unauthorized'),(402,'Payment Required'),(403,'Forbidden'),(404,'Not Found'),(405,'Method Not Allowed'),(406,'Not Acceptable'),(407,'Proxy Authentication Required'),(408,'Request Time-out'),(409,'Conflict'),(410,'Gone'),(411,'Length Required'),(412,'Precondition Failed'),(413,'Request Entity Too Large'),(414,'Request-URI Too Large'),(415,'Unsupported Media Type'),(416,'Requested range not satisfiable'),(417,'Expectation Failed'),(500,'Internal Server Error'),(501,'Not Implemented'),(502,'Bad Gateway'),(503,'Service Unavailable'),(504,'Gateway Time-out'),(505,'HTTP Version not supported');

#
# Filling agents table
#

INSERT INTO agent VALUES (1,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; DigExt)','Windows 98','Internet Explorer 5.0'),(2,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.0; Q312461)','Windows 2000','Internet Explorer 6.0'),(3,'Mozilla/4.0 (compatible; MSIE 5.13; Mac_PowerPC)','Mac OS','Internet Explorer 5.1'),(4,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Win 9x 4.90)','Windows 98','Internet Explorer 5.5'),(5,'',NULL,NULL),(6,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.0)','Windows 2000','Internet Explorer 6.0'),(7,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98)','Windows 98','Internet Explorer 6.0'),(8,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; DigExt; ZON)','Windows 98','Internet Explorer 5.0'),(9,'Mozilla/4.08 [en] (WinNT; I)','Windows NT','Netscape 4.08'),(10,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; TUCOWS)','Windows 98','Internet Explorer 5.5'),(11,'Mozilla/5.0 (Slurp/cat; slurp@inktomi.com; http://www.inktomi.com/slurp.html)',NULL,NULL),(12,'Mozilla/4.0 (compatible; MSIE 5.01; Windows NT 5.0)','Windows 2000','Internet Explorer 5.0'),(13,'Mozilla/4.0 (compatible; MSIE 5.5; Windows NT 5.0)','Windows 2000','Internet Explorer 5.5'),(14,'Mozilla/4.0 (compatible; MSIE 5.0; Mac_PowerPC)','Mac OS','Internet Explorer 5.0'),(15,'Mozilla/3.0 (Slurp/si; slurp@inktomi.com; http://www.inktomi.com/slurp.html)',NULL,NULL),(16,'Mozilla/4.0 (compatible; MSIE 5.5; Windows NT 4.0)','Windows NT 4.0','Internet Explorer 5.5'),(17,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 95; DigExt)','Windows 95','Internet Explorer 5.0'),(18,'Mozilla/4.0 (compatible; MSIE 4.01; Windows NT; Informatie Beheer Groep)','Windows NT 4.0','Internet Explorer 4.0'),(19,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1)','Windows XP','Internet Explorer 6.0'),(20,'FAST-WebCrawler/3.5 (atw-crawler at fast dot no; http://fast.no/support/crawler.asp)',NULL,'FAST-WebCrawler 3.5'),(21,'Mozilla/4.0 (compatible; MSIE 5.5; Windows NT 5.0) Fetch API Request','Windows 2000','Internet Explorer 5.5'),(22,'bumblebee@relevare.com',NULL,NULL),(23,'Mozilla/4.0 (compatible; MSIE 5.01; Windows 95; BLAN 2.0)','Windows 95','Internet Explorer 5.0'),(24,'lwp-request/1.39',NULL,NULL),(25,'Mercator-2.0',NULL,NULL),(26,'Mozilla/2.0 (compatible; Ask Jeeves)',NULL,'Ask Jeeves'),(27,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98)','Windows 98','Internet Explorer 5.5'),(28,'Mozilla/4.0 (compatible; MSIE 5.01; Windows 98)','Windows 98','Internet Explorer 5.0'),(29,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; {Tiscali})','Windows XP','Internet Explorer 6.0'),(30,'Mozilla/4.0 (compatible; MSIE 5.14; Mac_PowerPC)','Mac OS','Internet Explorer 5.1'),(31,'Mozilla/3.0 (Vagabondo/1.2 MT; webcrawler@NOSPAMwise-guys.nl; http://webagent.wise-guys.nl/)',NULL,NULL),(32,'Vagabondo/1.2 MT (webagent@NOSPAMwise-guys.nl)',NULL,NULL),(33,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Win 9x 4.90; Internet Explorer 5.0 - Sonera Internet version 1.1)','Windows 98','Internet Explorer 5.5'),(34,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; Win 9x 4.90; MSN 6.1; MSNbMSFT; MSNmnl-be; MSNc0z)','Windows 98','Internet Explorer 6.0'),(35,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; Win 9x 4.90; MSN 6.1; MSNbMSFT; MSNmnl-be; MSNc00)','Windows 98','Internet Explorer 6.0'),(36,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.0; T312461)','Windows 2000','Internet Explorer 6.0'),(37,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Wanadoo cable)','Windows 98','Internet Explorer 5.5'),(38,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; Q312461; .NET CLR 1.0.3705)','Windows 98','Internet Explorer 6.0'),(39,'Mozilla/4.0 (compatible; MSIE 5.5; Windows NT 4.0; Het Net 2.0)','Windows NT 4.0','Internet Explorer 5.5'),(40,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; T312461)','Windows 98','Internet Explorer 5.5'),(41,'Vagabondo/1.1 MT (webagent@NOSPAMwise-guys.nl)',NULL,NULL),(42,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Win 9x 4.90; Wanadoo cable)','Windows 98','Internet Explorer 5.5'),(43,'Mozilla/4.0 (compatible; MSIE 5.01; Windows 95)','Windows 95','Internet Explorer 5.0'),(44,'Mozilla/3.0 (compatible; Indy Library)',NULL,NULL),(45,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 95)','Windows 95','Internet Explorer 5.5'),(46,'Mozilla/4.0 (compatible; MSIE 5.0; CS 2000; Windows 98; DigExt)','Windows 98','Internet Explorer 5.0'),(47,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; PKBL008; RDC)','Windows 98','Internet Explorer 6.0'),(48,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 4.0)','Windows NT 4.0','Internet Explorer 6.0'),(49,'Mozilla/3.0',NULL,NULL),(50,'Slurp/si (slurp@inktomi.com; http://www.inktomi.com/slurp.html)',NULL,NULL),(51,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; telus.net_v5.0.1)','Windows 98','Internet Explorer 6.0'),(52,'libwww-perl/5.53',NULL,NULL),(53,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; Win 9x 4.90; Hotbar 4.0)','Windows 98','Internet Explorer 6.0'),(54,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; DigExt; {World Online})','Windows 98','Internet Explorer 5.0'),(55,'Mozilla/4.0 (compatible; MSIE 5.5; Windows NT 5.0; Wanadoo cable; Hotbar 3.0)','Windows 2000','Internet Explorer 5.5'),(56,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; BTinternet V8.4)','Windows 98','Internet Explorer 5.5'),(57,'Mozilla/4.0 (compatible; MSIE 5.0; Windows NT; DigExt)','Windows NT 4.0','Internet Explorer 5.0'),(58,'Mozilla/4.0 (compatible; MSIE 5.01; Windows NT; FD)','Windows NT 4.0','Internet Explorer 5.0'),(59,'Mozilla/4.0 (compatible; MSIE 5.0; CS 2000; Windows 95; DigExt; NOKNOK)','Windows 95','Internet Explorer 5.0'),(60,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; Q312461)','Windows XP','Internet Explorer 6.0'),(61,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; Wanadoo Cable 2.0c; Hotbar 3.0)','Windows XP','Internet Explorer 6.0'),(62,'ia_archiver',NULL,NULL),(63,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Wanadoo 5.3)','Windows 98','Internet Explorer 5.5'),(64,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; MSN 6.1; MSNbMSFT; MSNmit-it; MSNc0z)','Windows XP','Internet Explorer 6.0'),(65,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; MSN 6.1; MSNbMSFT; MSNmit-it; MSNc00)','Windows XP','Internet Explorer 6.0'),(66,'Mozilla/4.0 (compatible; Netcraft Web Server Survey)',NULL,NULL),(67,'Mozilla/4.0 (compatible; MSIE 5.5; Windows NT 4.0; DigExt)','Windows NT 4.0','Internet Explorer 5.5'),(68,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 95; HCCnet)','Windows 95','Internet Explorer 5.5'),(69,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.0; .NET CLR 1.0.3705)','Windows 2000','Internet Explorer 6.0'),(70,'Mozilla/4.51 [de]C-CCK-MCD DT  (Win98; I)','Windows 98','Netscape 4.51'),(71,'Mozilla/4.0 (compatible; Win32; WinHttp.WinHttpRequest.5)',NULL,NULL),(72,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.0; FreeSurf iweb.nl v1.0.2)','Windows 2000','Internet Explorer 6.0'),(73,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; Win 9x 4.90; Wanadoo NL; KITV4 Wanadoo)','Windows 98','Internet Explorer 6.0'),(74,'Mozilla/4.0 (compatible; MSIE 4.01; Windows 98; PKBL008; Wanadoo NL; KITV4 Wanadoo)','Windows 98','Internet Explorer 4.0'),(75,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Win 9x 4.90; MSOCD; AtHomeNL191)','Windows 98','Internet Explorer 5.5'),(76,'Mozilla/4.0 compatible ZyBorg/1.0 (ZyBorg@WISEnutbot.com; http://www.WISEnutbot.com)',NULL,NULL),(77,'Mozilla/4.0 (compatible; MSIE 4.01; Windows 95; Het Net 2.0)','Windows 95','Internet Explorer 4.0'),(78,'Mozilla/4.76 [en] (Win98; U)','Windows 98','Netscape 4.76'),(79,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Win 9x 4.90; CWIE55NL60; CHWIE_NL60)','Windows 98','Internet Explorer 5.5'),(80,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 95; DigExt; KITV4 Wanadoo)','Windows 95','Internet Explorer 5.0'),(81,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 95; DigExt; sureseeker.com)','Windows 95','Internet Explorer 5.0'),(82,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; CWIE55NL60; CHWIE_NL60)','Windows 98','Internet Explorer 6.0'),(83,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; Wanadoo cable; Q312461)','Windows XP','Internet Explorer 6.0'),(84,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; ZON; FreeSurf myweb.nl v1.0.1)','Windows 98','Internet Explorer 6.0'),(85,'Mozilla/4.75 [en] (Win95; U)','Windows 95','Netscape 4.75'),(86,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 95; Hotbar 3.0)','Windows 95','Internet Explorer 5.5'),(87,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 95; ZON)','Windows 95','Internet Explorer 5.5'),(88,'Microsoft URL Control - 6.00.8862','',NULL),(89,'Googlebot/2.1 (+http://www.googlebot.com/bot.html)',NULL,'Googlebot 2.1'),(90,'Slurp/si-emb (slurp@inktomi.com; http://www.inktomi.com/slurp.html)',NULL,NULL),(91,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; DigExt; FREELER1.1)','Windows 98','Internet Explorer 5.0'),(92,'Mozilla/4.0 (compatible; MSIE 5.01; Windows 98; HCCnet Internet Explorer)','Windows 98','Internet Explorer 5.0'),(93,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; T312461)','Windows 98','Internet Explorer 5.0'),(94,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; Q312461)','Windows 98','Internet Explorer 6.0'),(95,'Mozilla/4.0 (compatible; MSIE 5.01; MSN 2.5; Windows 98; {VIZZAVI0201})','Windows 98','Internet Explorer 5.0'),(96,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; Hotbar 3.0)','Windows XP','Internet Explorer 6.0'),(97,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; Win 9x 4.90)','Windows 98','Internet Explorer 6.0'),(98,'Mozilla/4.61 (Macintosh; I; PPC)','Mac OS','Netscape 4.61'),(99,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; Wanadoo cable; H010818)','Windows 98','Internet Explorer 5.0'),(100,'Mozilla/4.78 [fr] (Windows NT 5.0; U)','Windows 2000','Netscape 4.78'),(101,'Mozilla/4.0 (compatible; MSIE 5.5; Windows NT 5.0; T312461)','Windows 2000','Internet Explorer 5.5'),(102,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; DigExt; HCCnet Internet Explorer)','Windows 98','Internet Explorer 5.0'),(103,'Mozilla/4.0 (compatible; MSIE 5.01; Windows 98; BCD2000)','Windows 98','Internet Explorer 5.0'),(104,'Mozilla/4.79 [en] (Win98; U)','Windows 98','Netscape 4.79'),(105,'Mozilla/4.0 (compatible; MSIE 5.01; Windows NT 5.0; {World Online})','Windows 2000','Internet Explorer 5.0'),(106,'Openfind data gatherer, Openbot/3.0+(robot-response@openfind.com.tw;+http://www.openfind.com.tw/robot.html)',NULL,'Openbot 3.0'),(107,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Win 9x 4.90; FREELER1.2)','Windows 98','Internet Explorer 5.5'),(108,'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:0.9.4) Gecko/20010914','Linux','Mozilla 0.9'),(109,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; Win 9x 4.90; AtHomeNL191; Q312461)','Windows 98','Internet Explorer 6.0'),(110,'sitecheck.internetseer.com (For more info see: http://sitecheck.internetseer.com)',NULL,NULL),(111,'Mozilla/4.0 (compatible; MSIE 5.5; CS 2000; Windows 98; Win 9x 4.90; Het Net 4.0; H010818)','Windows 98','Internet Explorer 5.5'),(112,'Mozilla/4.72 (Macintosh; I; PPC)','Mac OS','Netscape 4.72'),(113,'Mozilla/4.0 (compatible; MSIE 5.01; Windows NT)','Windows NT 4.0','Internet Explorer 5.0'),(114,'Mozilla/3.01 (compatible;)',NULL,'Internet Explorer 3.0'),(115,'Mozilla/4.0 (compatible; MSIE 5.01; Windows 98; Freegates:youri; Hotbar 3.0)','Windows 98','Internet Explorer 5.0'),(116,'Mozilla/4.0 (compatible; MSIE 5.5; Windows NT 4.0; T312461; IE55232Ba)','Windows NT 4.0','Internet Explorer 5.5'),(117,'Mozilla/4.0+(compatible;+MSIE+6.0;)',NULL,NULL),(118,'Mozilla/4.0 (compatible; MSIE 4.5; Mac_PowerPC)','Mac OS','Internet Explorer 4.5'),(119,'Mozilla/4.0 (compatible; MSIE 5.5; Windows NT 5.0; H010818; T312461)','Windows 2000','Internet Explorer 5.5'),(120,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; DigExt; Hotbar 3.0)','Windows 98','Internet Explorer 5.0'),(121,'Mozilla/4.0 (compatible; MSIE 5.01; Windows 95; Wegener)','Windows 95','Internet Explorer 5.0'),(122,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; .NET CLR 1.0.3705)','Windows XP','Internet Explorer 6.0'),(123,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; DigExt; Wanadoo NL)','Windows 98','Internet Explorer 5.0'),(124,'Mozilla/5.0 (Macintosh; U; PPC Mac OS X; en-US; rv:1.0.0+) Gecko/20020423','Mac OS X','Mozilla 1.0'),(125,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Het Net 4.0)','Windows 98','Internet Explorer 5.5'),(126,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; ZON)','Windows 98','Internet Explorer 6.0'),(127,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; Hotbar 3.0)','Windows 98','Internet Explorer 6.0'),(128,'Mozilla/4.0_(compatible;_MSIE_5.0;_Windows_95)_VoilaBot/1.6 libwww/5.3.2','Windows 95','Internet Explorer 5.0'),(129,'Mozilla/4.73C-CCK-MCD {C-UDP; EBM-APPLE} (Macintosh; U; PPC)','Mac OS','Netscape 4.73'),(130,'Mozilla/4.73 [de]C-CCK-MCD DT  (Win98; U)','Windows 98','Netscape 4.73'),(131,'Mozilla/4.0 (compatible; MSIE 5.5; Windows NT 4.0; TUCOWS.COM)','Windows NT 4.0','Internet Explorer 5.5'),(132,'Xenu Link Sleuth 1.2b',NULL,NULL),(133,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; DigExt; Het Net 3.2)','Windows 98','Internet Explorer 5.0'),(134,'Mozilla/4.0 (compatible; MSIE 5.01; Windows 98; FreeSurf myweb.nl v1.0.1)','Windows 98','Internet Explorer 5.0'),(135,'Mozilla/4.77C-CCK-MCD {C-UDP; EBM-APPLE} (Macintosh; U; PPC)','Mac OS','Netscape 4.77'),(136,'Microsoft URL Control - 6.00.8169',NULL,NULL),(137,'Scooter-W3.1.2',NULL,'Scooter 3.1'),(138,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; DigExt)','Windows 98','Internet Explorer 5.5'),(139,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Win 9x 4.90; Het Net 4.0)','Windows 98','Internet Explorer 5.5'),(140,'Mozilla/4.51 [nl] (Win98; I)','Windows 98','Netscape 4.51'),(141,'Mozilla/4.7 [nl] (Win98; I)','Windows 98','Netscape 4.7'),(142,'Mozilla/4.0 (compatible; MSIE 5.01; Windows NT 5.0; Het Net 3.2)','Windows 2000','Internet Explorer 5.0'),(143,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 4.0; T312461)','Windows NT 4.0','Internet Explorer 6.0'),(144,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; CWIE55NL60; CHWIE_NL60)','Windows 98','Internet Explorer 5.5'),(145,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.0; T312461; .NET CLR 1.0.3705)','Windows 2000','Internet Explorer 6.0'),(146,'Mozilla/4.0 (compatible; MSIE 5.5; Windows NT 4.0; T312461)','Windows NT 4.0','Internet Explorer 5.5'),(147,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 95; DigExt; T312461)','Windows 95','Internet Explorer 5.5'),(148,'Mozilla/4.0 (compatible; MSIE 4.01; Windows 95)','Windows 95','Internet Explorer 4.0'),(149,'Mozilla/4.0 (compatible; MSIE 4.01; MSN 2.5; MSN 2.5; Windows 95)','Windows 95','Internet Explorer 4.0'),(150,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.0; DigExt; Tucows)','Windows 2000','Internet Explorer 6.0'),(151,'Mozilla/4.0 (compatible; MSIE 4.01; Windows NT)','Windows NT 4.0','Internet Explorer 4.0'),(152,'http://www.almaden.ibm.com/cs/crawler',NULL,NULL),(153,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.0; TUCOWS; Q312461)','Windows 2000','Internet Explorer 6.0'),(154,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Hotbar 3.0)','Windows 98','Internet Explorer 5.5'),(155,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Cable Wanadoo)','Windows 98','Internet Explorer 5.5'),(156,'Mozilla/4.0 (compatible; MSIE 5.01; Windows NT 5.0; Demon NL v9906.03)','Windows 2000','Internet Explorer 5.0'),(157,'Mozilla/4.72 [en] (WinNT; U)','Windows NT','Netscape 4.72'),(158,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Creative)','Windows 98','Internet Explorer 5.5'),(159,'Mozilla/4.0 (compatible; MSIE 5.01; Windows 98; Powered by Uni-One Internet (Free))','Windows 98','Internet Explorer 5.0'),(160,'Mozilla/4.06 [nl] (Win95; I)','Windows 95','Netscape 4.06'),(161,'Mozilla/4.0 (compatible; MSIE 5.01; CS 2000; Windows 98)','Windows 98','Internet Explorer 5.0'),(162,'ilse/1.0',NULL,NULL),(163,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; H010818)','Windows 98','Internet Explorer 5.5'),(164,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; DigExt; KA V3 NL)','Windows 98','Internet Explorer 5.0'),(165,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; T312461; Crazy Browser 1.0.5)','Windows 98','Internet Explorer 5.5'),(166,'SmartDownload/1.2.77 (Win32; Feb  1 2000)',NULL,'SmartDownload 1.2'),(167,'Mozilla/4.0 (compatible; MSIE 5.01; Windows 98; ZON)','Windows 98','Internet Explorer 5.0'),(168,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; Wanadoo cable; Q312461)','Windows 98','Internet Explorer 6.0'),(169,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; MSOCD)','Windows 98','Internet Explorer 6.0'),(170,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; PKBL008; DigExt)','Windows 98','Internet Explorer 5.0'),(171,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; FREELER1.1)','Windows 98','Internet Explorer 5.5'),(172,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; Win 9x 4.90; Q312461)','Windows 98','Internet Explorer 6.0'),(173,'Mozilla/4.0 (compatible; MSIE 5.01; Windows 98; Creative)','Windows 98','Internet Explorer 5.0'),(174,'Mozilla/4.0 (compatible; MSIE 5.12; Mac_PowerPC)','Mac OS','Internet Explorer 5.1'),(175,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; WI4C 1.1.0.515)','Windows 98','Internet Explorer 5.5'),(176,'Mozilla/4.0 (compatible; MSIE 4.01; Windows 98)','Windows 98','Internet Explorer 4.0'),(177,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 95; DigExt; FREELER1.1)','Windows 95','Internet Explorer 5.0'),(178,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; PKBL008; DigExt; {World Online}; Hotbar 3.0)','Windows 98','Internet Explorer 5.0'),(179,'Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US; rv:0.9.4) Gecko/20011128 Netscape6/6.2.1','Windows XP','Netscape 6.2'),(180,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; {World Online - Abn Amro}; Q312461)','Windows 98','Internet Explorer 6.0'),(181,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; FreeSurf myweb.nl v1.0.1)','Windows XP','Internet Explorer 6.0'),(182,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; WISH-01-01)','Windows 98','Internet Explorer 5.5'),(183,'Mozilla/2.0 (compatible; MSIE 3.02; Update a; AK; Windows 95)','Windows 95','Internet Explorer 3.0'),(184,'Mozilla/5.0 (Windows; U; Win98; en-US; rv:1.0rc1) Gecko/20020417','Windows 98','Mozilla 1.0'),(185,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; DigExt; ZeelandNet TeleFree V.1)','Windows 98','Internet Explorer 5.0'),(186,'Mozilla/4.0 (compatible; MSIE 5.01; Windows 98; 24hours01)','Windows 98','Internet Explorer 5.0'),(187,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; Q312461; .NET CLR 1.0.3705)','Windows XP','Internet Explorer 6.0'),(188,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Win 9x 4.90; Cablewanadoo basic)','Windows 98','Internet Explorer 5.5'),(189,'Scooter-3.2',NULL,'Scooter 3.2'),(190,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; DigExt; AtHomeNL0107)','Windows 98','Internet Explorer 6.0'),(191,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Win 9x 4.90; H010818)','Windows 98','Internet Explorer 5.5'),(192,'Mozilla/4.0 (compatible; MSIE 5.01; Windows 98; {World Online})','Windows 98','Internet Explorer 5.0'),(193,'Mozilla/4.78 [en]C-CCK-MCD   (Windows NT 5.0; U)','Windows 2000','Netscape 4.78'),(194,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.0; H010818)','Windows 2000','Internet Explorer 6.0'),(195,'curl/7.9 (sparc-sun-solaris2.7) libcurl 7.9','Solaris','curl 7.9'),(196,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.0; T312461; Q312461; .NET CLR 1.0.3705)','Windows 2000','Internet Explorer 6.0'),(197,'Mozilla/4.78 (Macintosh; U; PPC)','Mac OS','Netscape 4.78'),(198,'Mozilla/4.0 (compatible; BorderManager 3.0)',NULL,NULL),(199,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Win 9x 4.90; H010818; {Vizzavi052001})','Windows 98','Internet Explorer 5.5'),(200,'Mozilla/4.0 (compatible; MSIE 5.01; Windows NT 5.0; FreeSurf myweb.nl v1.0.1)','Windows 2000','Internet Explorer 5.0'),(201,'Mozilla/4.0 (compatible; MSIE 5.01; Windows 95; NCRV-Net)','Windows 95','Internet Explorer 5.0'),(202,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; Wanadoo Cable 2.0c)','Windows XP','Internet Explorer 6.0'),(203,'Mozilla/4.0 (compatible; MSIE 5.0; Windows NT)','Windows NT 4.0','Internet Explorer 5.0'),(204,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 95; T312461)','Windows 95','Internet Explorer 5.5'),(205,'Mozilla/4.0 (compatible; MSIE 5.5; Windows NT 5.0; DigExt)','Windows 2000','Internet Explorer 5.5'),(206,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; NRCnet Internet)','Windows 98','Internet Explorer 6.0'),(207,'Mozilla/4.0 (compatible; MSIE 5.01; Windows NT; MSNCOM)','Windows NT 4.0','Internet Explorer 5.0'),(208,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; sureseeker.com)','Windows 98','Internet Explorer 5.0'),(209,'Mozilla/4.08 [en] (Win98; I)','Windows 98','Netscape 4.08'),(210,'Mozilla/4.0 (compatible; MSIE 5.5; Windows NT 4.0; H010818)','Windows NT 4.0','Internet Explorer 5.5'),(211,'BaiDuSpider',NULL,NULL),(212,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; Wanadoo 5.4)','Windows XP','Internet Explorer 6.0'),(213,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Win 9x 4.90; HCCnet Internet Explorer)','Windows 98','Internet Explorer 5.5'),(214,'Mozilla/4.75 [en] (Win98; U)','Windows 98','Netscape 4.75'),(215,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 95; DigExt; MSNCOM)','Windows 95','Internet Explorer 5.0'),(216,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; MSOCD; AtHomeEN191)','Windows 98','Internet Explorer 5.5'),(217,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; {Tiscali}; MSN 6.1; MSNbMSFT; MSNmnl-nl; MSNcOTH)','Windows XP','Internet Explorer 6.0'),(218,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; Het Net 4.0)','Windows XP','Internet Explorer 6.0'),(219,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 95; Wanadoo cable)','Windows 95','Internet Explorer 5.5'),(220,'Webspider 1.0b8',NULL,NULL),(221,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; DigExt; UUNET Nederland)','Windows 98','Internet Explorer 5.0'),(222,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Win 9x 4.90; KITV4 Wanadoo)','Windows 98','Internet Explorer 5.5'),(223,'Mozilla/4.7C-CCK-MCD  (Macintosh; I; PPC)','Mac OS','Netscape 4.7'),(224,'Mozilla/4.0 (compatible; MSIE 5.5; Windows NT 5.0; LLY-EMA-EN)','Windows 2000','Internet Explorer 5.5'),(225,'Mozilla/4.75 (Macintosh; U; PPC)','Mac OS','Netscape 4.75'),(226,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; MSNCOM)','Windows 98','Internet Explorer 5.5'),(227,'Mozilla/4.0 (compatible; MSIE 4.0; Windows NT)','Windows NT 4.0','Internet Explorer 4.0'),(228,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; DigExt; TiscaliFreeNet)','Windows 98','Internet Explorer 5.0'),(229,'IP*Works! V5 HTTP/S Component - by /n software - www.nsoftware.com',NULL,NULL),(230,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; FreeSurf myweb.nl v1.0.1)','Windows 98','Internet Explorer 6.0'),(231,'Mozilla/4.0 (compatible; MSIE 5.01; Windows 98; BT Ignite)','Windows 98','Internet Explorer 5.0'),(232,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; Q312461; Linux gnu. debian of zow)','Windows XP','Internet Explorer 6.0'),(233,'WinampMPEG/2.7',NULL,'Winamp 2.7'),(234,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; Win 9x 4.90; Hotbar 3.0)','Windows 98','Internet Explorer 6.0'),(235,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; DigExt)','Windows XP','Internet Explorer 6.0'),(236,'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:0.9.4) Gecko/20011126 Netscape6/6.2.1','Linux','Netscape 6.2'),(237,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.0; NCRV-Net; Q312461)','Windows 2000','Internet Explorer 6.0'),(238,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; {World Online})','Windows 98','Internet Explorer 6.0'),(239,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; Wanadoo NL; KITV4 Wanadoo)','Windows 98','Internet Explorer 6.0'),(240,'Mozilla/4.51 [nl] (Win95; I)','Windows 95','Netscape 4.51'),(241,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Win 9x 4.90; Wanadoo cable; H010818)','Windows 98','Internet Explorer 5.5'),(242,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 95; Wanadoo cable; H010818)','Windows 95','Internet Explorer 5.5'),(243,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; Q312461; CHWIE_NL60)','Windows 98','Internet Explorer 6.0'),(244,'Mozilla/4.0 (compatible; MSIE 5.01; Windows 95; SKY11a)','Windows 95','Internet Explorer 5.0'),(245,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; DigExt; Hi Way, powered by Planet Internet)','Windows 98','Internet Explorer 5.0'),(246,'contype',NULL,NULL),(247,'Mozilla/5.0 (Macintosh; U; PPC Mac OS X; en-US; rv:0.9.4.1) Gecko/20020315 Netscape6/6.2.2','Mac OS X','Netscape 6.2'),(248,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 95; FreeSurf myweb.nl v1.0.1)','Windows 95','Internet Explorer 5.5'),(249,'Mozilla/4.0 (compatible; MSIE 5.01; Windows 98; TNET5.0NL)','Windows 98','Internet Explorer 5.0'),(250,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; DigExt; {World Online - Abn Amro})','Windows 98','Internet Explorer 5.0'),(251,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; Q312461; Crazy Browser 1.0.5)','Windows 98','Internet Explorer 6.0'),(252,'LinkWalker',NULL,''),(253,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98)','Windows 98','Internet Explorer 5.0'),(254,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; DigExt; FA V2 NL)','Windows 98','Internet Explorer 5.0'),(255,'ScoutAbout',NULL,NULL),(256,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; MSOCD; AtHomeNL191)','Windows 98','Internet Explorer 6.0'),(257,'Mozilla/4.0 (compatible; MSIE 5.01; Windows 98; Het Net 3.0)','Windows 98','Internet Explorer 5.0'),(258,'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:0.9.2.1) Gecko/20010901','Linux','Mozilla 0.9'),(259,'Mozilla/4.0 (compatible; MSIE 5.01; Windows NT 5.0; Hotbar 3.0)','Windows 2000','Internet Explorer 5.0'),(260,'DSurf15a 31',NULL,NULL),(261,'Mozilla/4.73 (Macintosh; I; PPC)','Mac OS','Netscape 4.73'),(262,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; Het Net 3.2)','Windows XP','Internet Explorer 6.0'),(263,'Mozilla/4.0 (compatible ; MSIE 6.0; Windows NT 5.1)','Windows XP','Internet Explorer 6.0'),(264,'Scooter/1.0',NULL,'Scooter 1.0'),(265,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 95; InterNeeds.net; DigExt)','Windows 95','Internet Explorer 5.0'),(266,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; {World Online})','Windows 98','Internet Explorer 5.5'),(267,'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:0.9.9) Gecko/20020507','Linux','Mozilla 0.9'),(268,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98) Opera 6.01  [en]','Windows 98','Opera 6.0'),(269,'Mozilla/4.0 (compatible; MSIE 6.0b; Windows 98)','Windows 98','Internet Explorer 6.0'),(270,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; Wanadoo cable)','Windows 98','Internet Explorer 6.0'),(271,'Arachnia 1.3  (Spider 2.1/info@vang.net)',NULL,NULL),(272,'Mozilla/4.0 (compatible; MSIE 5.0; Windows NT; DigExt; INFRA2000)','Windows NT','Internet Explorer 5.0'),(273,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 95; Hotbar 2.0)','Windows 98','Internet Explorer 5.5'),(274,'Mozilla/4.5 (compatible; OmniWeb/4.1-v399; Mac_PowerPC)','Mac OS','OmniWeb 4.1'),(275,'Mozilla/4.79 (Macintosh; U; PPC)','Mac OS','Netscape 4.79'),(276,'Scooter-3.2.EX',NULL,NULL),(277,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; DigExt; Wanadoo BE; KITV4 Wanadoo)','Windows 98','Internet Explorer 5.0'),(278,'Mozilla/4.0 (compatible; MSIE 4.01; AOL 5.0; Windows 95)','Windows 98','Internet Explorer 4.01'),(279,'Googlebot-Image/1.0 (+http://www.googlebot.com/bot.html)',NULL,NULL),(280,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Win 9x 4.90; Hotbar 3.0)','Windows 98','Internet Explorer 5.5'),(281,'Mozilla/4.0 (compatible; MSIE 5.0; Mac_PowerPC) Opera 5.0  [en]','Mac OS','Internet Explorer 5.0'),(282,'Mozilla/4.71 [en] (Win98; I)','Windows 98','Netscape 4.71'),(283,'Mozilla/5.0 (Macintosh; U; PPC; en-US; rv:0.9.4.1) Gecko/20020318 Netscape6/6.2.2','Mac OS','Netscape 6.2'),(284,'Mozilla/4.7 (Macintosh; I; PPC)','Mac OS','Netscape 4.7'),(285,'Mozilla/3.0 (compatible)',NULL,NULL),(286,'ASPSeek/1.2.7',NULL,NULL),(287,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; TUCOWS)','Windows XP','Internet Explorer 6.0'),(288,'Bilbo/2.3b-UNIX','Unix',NULL),(289,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; MSOCD; AtHomeNL191)','Windows 98','Internet Explorer 5.5'),(290,'Mozilla/4.0 (compatible; MSIE 5.5; Windows NT 5.0) WebWasher 3.0','Windows 2000','Internet Explorer 5.5'),(291,'Mozilla/4.75C-CCK-MCD {C-UDP; EBM-APPLE} (Macintosh; U; PPC)','Mac OS','Netscape 4.75'),(292,'Mozilla/4.0 (compatible; MSIE 4.01; Windows 95; Wegener)','Windows 98','Internet Explorer 4.01'),(293,'WebCopier v3.0.1',NULL,NULL),(294,'Mozilla/4.0 (compatible; MSIE 5.01; Windows NT; TUCOWS Network)','Windows NT','Internet Explorer 5.0'),(295,'Mozilla/5.0 (X11; U; Linux 2.4.4 i686)','Linux',NULL),(296,'Mozilla/4.0 (compatible; MSIE 5.5; CP/M; 64-bit high_encr)',NULL,'Internet Explorer 5.5'),(297,'Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US; rv:0.9.4.1) Gecko/20020314 Netscape6/6.2.2','Windows XP','Netscape 6.2'),(298,'Mozilla/4.7C-CCK-MCD {C-UDP; EBM-APPLE} (Macintosh; I; PPC)','Mac OS','Netscape 4.7'),(299,'Mozilla/4.0 (compatible; MSIE 5.5; MSN 2.5; Windows 98; T312461)','Windows 98','Internet Explorer 5.5'),(300,'FAST-WebCrawler/3.5 (atw-crawler at fast dot no; http://fast.no/support.php?c=faqs/crawler)',NULL,NULL),(301,'Mozilla/4.0 (compatible; MSIE 5.5; CS 2000; Windows 98; Win 9x 4.90)','Windows 98','Internet Explorer 5.5'),(302,'Mozilla/4.0 (compatible; MSIE 4.5; Mac_PowerPC; TECH373tnet_45m)','Mac OS','Internet Explorer 4.5'),(303,'Mozilla/4.7 [de]C-CCK-MCD DT  (Win95; I)','Windows 95','Netscape 4.7'),(304,'Mozilla/4.0 (compatible; MSIE 5.0; Windows ME) Opera 6.01  [ja]','Windows ME','Internet Explorer 5.0'),(305,'Mozilla/4.0 (compatible; MSIE 5.5; Windows NT 5.0; MSOCD; AtHomeNL191)','Windows 2000','Internet Explorer 5.5'),(306,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Win 9x 4.90; Het Net 3.2)','Windows 98','Internet Explorer 5.5'),(307,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Win 9x 4.90; Wanadoo NL; KITV4 Wanadoo)','Windows 98','Internet Explorer 5.5'),(308,'Mozilla/4.0 (compatible; MSIE 5.01; Windows 98; WanadooNL)','Windows 98','Internet Explorer 5.0'),(309,'INGRID/1.0',NULL,NULL),(310,'Mozilla/4.0 (compatible; MSIE 4.01; MSN 2.6; Windows 95)','Windows 98','Internet Explorer 4.01'),(311,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; T312461; Q312461)','Windows 98','Internet Explorer 6.0'),(312,'Mozilla/4.6 [en]C-SYMPA  (Win98; U)','Windows 98','Netscape 4.6'),(313,'vb wininet',NULL,NULL),(314,'Mozilla/4.76 [en] (WinNT; U)','Windows NT','Netscape 4.76'),(315,'Mozilla/2.0 (compatible; MS FrontPage 3.0)',NULL,NULL),(316,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; DigExt; Het Net 3.0)','Windows 98','Internet Explorer 5.0'),(317,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; AtHomeNL0107)','Windows 98','Internet Explorer 6.0'),(318,'Mozilla/4.0 (compatible; MSIE 5.01; Windows NT; KPN_MOL)','Windows NT','Internet Explorer 5.0'),(319,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 95; N_o_k_i_a; N o k i a)','Windows 98','Internet Explorer 5.5'),(320,'Lynx/2.8.4dev.16 libwww-FM/2.14 SSL-MM/1.4.1 OpenSSL/0.9.6',NULL,'Lynx 2.8'),(321,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.0; T312461; Q312461)','Windows 2000','Internet Explorer 6.0'),(322,'Mozilla/4.0 (compatible; MSIE 4.01; Windows 95; Joods Historisch Museum)','Windows 98','Internet Explorer 4.01'),(323,'Mozilla/3.0 (Vagabondo/1.1 MT; webcrawler@NOSPAMwise-guys.nl; http://webagent.wise-guys.nl/)',NULL,NULL),(324,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; Het Net 4.0)','Windows 98','Internet Explorer 6.0'),(325,'Mozilla/5.0 (Windows; U; Win98; en-US; rv:0.9.4) Gecko/20011128 Netscape6/6.2.1','Windows 98','Netscape 6.2'),(326,'Vagabondo/1.1 MT (webagent@wise-guys.nl)',NULL,NULL),(327,'Mozilla/4.0 (compatible; MSIE 4.01; Mac_PowerPC)','Mac OS','Internet Explorer 4.01'),(328,'Mozilla/4.0 (compatible; MSIE 5.0; Windows NT; DigExt; Centraal Boekhuis Culemborg, The Netherlands)','Windows NT','Internet Explorer 5.0'),(329,'Mozilla/4.0 (compatible; MSIE 5.5; Windows NT 5.0; Het Net 3.2)','Windows 2000','Internet Explorer 5.5'),(330,'Mozilla/4.5 (compatible; OmniWeb/4.0.6; Mac_PowerPC)','Mac OS','OmniWeb 4.0.6'),(331,'Mozilla/4.51 [zh-tw] (Win98; I)','Windows 98','Netscape 4.51'),(332,'Mozilla/4.0 (compatible; MSIE 5.0; Mac_PowerPC; S425166QXM0330e)','Mac OS','Internet Explorer 5.0'),(333,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; Cable Wanadoo; Hotbar 3.0)','Windows 98','Internet Explorer 6.0'),(334,'Mozilla/4.0 (compatible; MSIE 5.01; Windows 98; SPINWAY.COM)','Windows 98','Internet Explorer 5.0'),(335,'Mozilla/4.0 (compatible; MSIE 5.01; Windows 98; wish-00-07)','Windows 98','Internet Explorer 5.0'),(336,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; ZON; DigExt)','Windows 98','Internet Explorer 5.0'),(337,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; DigExt; Q312461)','Windows 98','Internet Explorer 6.0'),(338,'Mozilla/4.5 [en] (Win95; I)','Windows 95','Netscape 4.5'),(339,'Mozilla/4.0 (compatible; MSIE 4.01; CS 2000; Windows 98)','Windows 98','Internet Explorer 4.01'),(340,'Mozilla/4.5 (compatible; OmniWeb/4.1-v380; Mac_PowerPC)','Mac OS','OmniWeb 4.1'),(341,'Mozilla/4.75 [en] (X11; U; Linux 2.2.17 i686)','Linux','Netscape 4.75'),(342,'Mozilla/3.01-C-MACOS8 (Macintosh; I; PPC)','Mac OS',NULL),(343,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; FREELER1)','Windows 98','Internet Explorer 6.0'),(344,'Mozilla/4.0 (compatible; MSIE 5.01; Windows 98; Het Net 3.2)','Windows 98','Internet Explorer 5.0'),(345,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 95; Cable Wanadoo)','Windows 98','Internet Explorer 5.5'),(346,'Mozilla/2.0 (compatible; T-H-U-N-D-E-R-S-T-O-N-E)',NULL,NULL),(347,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; MSNCOM)','Windows XP','Internet Explorer 6.0'),(348,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; D-TNT 6 Beta)','Windows 98','Internet Explorer 6.0'),(349,'Mozilla/4.0 (compatible; MSIE 5.01; Windows NT; D_O_B)','Windows NT','Internet Explorer 5.0'),(350,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; HCCnet Internet Explorer)','Windows 98','Internet Explorer 5.5'),(351,'Scooter-3.2.NIV',NULL,NULL),(352,'ASPseek/1.2.9pre',NULL,NULL),(353,'Mozilla/3.01C-WXS-16  (Win95; I)','Windows 95',NULL),(354,'Mozilla/4.0 (compatible; MSIE 5.5; Windows NT 5.0; H010818)','Windows 2000','Internet Explorer 5.5'),(355,'Mozilla/4.0 (compatible; MSIE 5.01; Windows 98; Solcon ISP)','Windows 98','Internet Explorer 5.0'),(356,'Mozilla/4.0 (compatible; MSIE 5.0; Windows NT; DigExt; Gemeente Leiden)','Windows NT','Internet Explorer 5.0'),(357,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; {World Online}; MSN 6.1; MSNbMSFT; MSNmnl-nl; MSNc0z)','Windows 98','Internet Explorer 5.5'),(358,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; {World Online}; MSN 6.1; MSNbMSFT; MSNmnl-nl; MSNc00)','Windows 98','Internet Explorer 5.5'),(359,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; {World Online}; MSN 6.1; MSNbMSFT; MSNmnl-nl; MSNc00), Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; {World Online}; MSN 6.1; MSNbMSFT; MSNmnl-nl; MSNc00)','Windows 98','Internet Explorer 5.5'),(360,'Mozilla/5.0 (Macintosh; U; PPC Mac OS X; en-US; rv:0.9.4) Gecko/20011126 Netscape6/6.2.1','Mac OS X','Netscape 6.2'),(361,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.0; matlas-2.0.2145)','Windows 2000','Internet Explorer 6.0'),(362,'URL_Access/2.1 (Macintosh; PPC)','Mac OS',NULL),(363,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Win 9x 4.90; {VIZZAVI0101})','Windows 98','Internet Explorer 5.5'),(364,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Wanadoo cable; H010818)','Windows 98','Internet Explorer 5.5'),(365,'Mozilla/4.76 (Macintosh; U; PPC)','Mac OS','Netscape 4.76'),(366,'Wget/1.6',NULL,'Wget 1.6'),(367,'URL_Access/2.0 (Macintosh; PPC)','Mac OS',NULL),(368,'Python-urllib/1.10',NULL,NULL),(369,'Mozilla/5.0',NULL,NULL),(370,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; Win 9x 4.90; TNET5.0NL)','Windows 98','Internet Explorer 6.0'),(371,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; Win 9x 4.90; Telfort Online 1.0)','Windows 98','Internet Explorer 6.0'),(372,'Mozilla/4.71 (Macintosh; I; PPC)','Mac OS','Netscape 4.71'),(373,'Mozilla/2.0 (compatible; NEWT ActiveX; Win32)',NULL,NULL),(374,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.0; AIRF; Q312461)','Windows 2000','Internet Explorer 6.0'),(375,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 4.0; Weberbank Browser V1.0)','Windows NT 4.0','Internet Explorer 6.0'),(376,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; DigExt; Demon NL v9906.03)','Windows 98','Internet Explorer 5.0'),(377,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; DigExt; MSNCOM)','Windows 98','Internet Explorer 5.0'),(378,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; FreeSurf etrade.nl v1.0.1)','Windows 98','Internet Explorer 5.5'),(379,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; {World Online})','Windows XP','Internet Explorer 6.0'),(380,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; Planet-ADSL)','Windows 98','Internet Explorer 5.0'),(381,'Mozilla/3.0 (Vagabondo/1.1 MT; webagent@wise-guys.nl; http://webagent.wise-guys.nl/)',NULL,NULL),(382,'Mozilla/5.0 (Windows; U; Win98; en-US; rv:0.9.2) Gecko/20010726 Netscape6/6.1','Windows 98','Netscape 6.1'),(383,'Mozilla/4.0 (compatible; MSIE 5.01; Windows NT 4.0)','Windows NT 4.0','Internet Explorer 5.0'),(384,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; DigExt; Creative)','Windows 98','Internet Explorer 5.0'),(385,'Mozilla/4.0 (compatible; MSIE 4.0; Windows 95; Hotbar 2.0)','Windows 98','Internet Explorer 4.0'),(386,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; {Vizzavi052001}; Q312461)','Windows 98','Internet Explorer 6.0'),(387,'Mozilla/4.0 (compatible; MSIE 5.01; Windows 95; VNIE5 4.0 (901))','Windows 98','Internet Explorer 5.0'),(388,'Mozilla/4.77 (Macintosh; U; PPC)','Mac OS','Netscape 4.77'),(389,'Mozilla/4.0 (compatible; MSIE 5.5; Windows NT 5.0; LNV)','Windows 2000','Internet Explorer 5.5'),(390,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; NOKNOK; Q312461)','Windows 98','Internet Explorer 6.0'),(391,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 95; DigExt; ZON)','Windows 98','Internet Explorer 5.0'),(392,'Internet Explorer','Windows',NULL),(393,'Links SQL (http://gossamer-threads.com/scripts/links-sql/)',NULL,NULL),(394,'Mozilla/4.0 (compatible; MSIE 6.0b; Windows NT 5.0)','Windows 2000','Internet Explorer 6.0'),(395,'Mozilla/4.0 (compatible; MSIE 5.5; Windows NT 5.0; HEWLETT-PACKARD -- IE 5.5 SP1; H010818)','Windows 2000','Internet Explorer 5.5'),(396,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; DigExt; FreeSurf tmfweb.nl v1.0.1)','Windows 98','Internet Explorer 5.0'),(397,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; c\'t)','Windows 98','Internet Explorer 6.0'),(398,'Mozilla/4.5 [nl]C-CCK-MCD   (Win98; I)','Windows 98','Netscape 4.5'),(399,'Mozilla/4.0 (compatible; MSIE 5.5; Windows NT 4.0; LARENSTEIN)','Windows NT 4.0','Internet Explorer 5.5'),(400,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; Win 9x 4.90; WanadooNL40)','Windows 98','Internet Explorer 6.0'),(401,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; CWIE55NL60; CHWIE_NL60; Q312461)','Windows 98','Internet Explorer 6.0'),(402,'Mozilla/4.0 (compatible; MSIE 5.0; Windows NT; DigExt; TUCOWS)','Windows NT','Internet Explorer 5.0'),(403,'Mozilla/4.77 [en] (WinNT; U)','Windows NT','Netscape 4.77'),(404,'Mozilla/4.6 [en] (Win98; I)','Windows 98','Netscape 4.6'),(405,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; DigExt; T312461)','Windows 98','Internet Explorer 5.5'),(406,'Mozilla/4.0 (compatible; MSIE 5.0; Mac_PowerPC; TECH373tnet_45m)','Mac OS','Internet Explorer 5.0'),(407,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; DigExt; libero)','Windows 98','Internet Explorer 5.0'),(408,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; Win 9x 4.90; FREELER1.2)','Windows 98','Internet Explorer 6.0'),(409,'Mozilla/5.0 (Macintosh; U; PPC; en-US; rv:0.9.4) Gecko/20011130 Netscape6/6.2.1','Mac OS','Netscape 6.2'),(410,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; DigExt; DVD Owner)','Windows XP','Internet Explorer 6.0'),(411,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98) Opera 6.01  [nl]','Windows 98','Internet Explorer 5.0'),(412,'Mozilla/4.7 [en]C-CCK-MCD NSCPCD47  (WinNT; I)','Windows NT','Netscape 4.7'),(413,'Nutscrape/1.0 (CP/M; 8-bit)',NULL,NULL),(414,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Win 9x 4.90; CWIE55NL60; CHWIE_NL60; H010818)','Windows 98','Internet Explorer 5.5'),(415,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; PKBL008; DigExt; ZON)','Windows 98','Internet Explorer 5.0'),(416,'Mozilla/4.7 [nl]C-CCK-MCD SURFKITV4  (Win98; U)','Windows 98','Netscape 4.7'),(417,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; PKBL008; {World Online})','Windows 98','Internet Explorer 6.0'),(418,'MSProxy/2.0',NULL,NULL),(419,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; DigExt; FreeSurf myweb.nl v1.0.1)','Windows 98','Internet Explorer 5.0'),(420,'Mozilla/5.0 (X11; U; Linux i586; en-US; rv:0.9.5) Gecko/20011012','Linux','Mozilla 0.9'),(421,'Mozilla/4.0 (compatible; MSIE 4.01; Windows 98; Wanadoo NL)','Windows 98','Internet Explorer 4.01'),(422,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Wanadoo cable; Hotbar 3.0)','Windows 98','Internet Explorer 5.5'),(423,'Mozilla/4.51 [en] (WinNT; I)','Windows NT','Netscape 4.51'),(424,'Mozilla/4.5 [de] (Macintosh; I; PPC)','Mac OS','Netscape 4.5'),(425,'Mozilla/4.0 (compatible; MSIE 4.01; Windows 95; Wanadoo NL; KITV4 Wanadoo)','Windows 98','Internet Explorer 4.01'),(426,'Mozilla/4.7 [en] (Win95; I)','Windows 95','Netscape 4.7'),(427,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; Tucows)','Windows 98','Internet Explorer 6.0'),(428,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; DigExt; AgroWeb Internet)','Windows 98','Internet Explorer 5.0'),(429,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; Wanadoo cable)','Windows XP','Internet Explorer 6.0'),(430,'Mozilla/4.7 (compatible; http://eidetica.com/spider)',NULL,'Netscape 4.7'),(431,'Mozilla/4.0 (compatible; MSIE 4.01; Windows 98; Planet Internet)','Windows 98','Internet Explorer 4.01'),(432,'Mozilla/4.0 (compatible; MSIE 5.5; Windows NT 5.0; FreeSurf myweb.nl v1.0.1)','Windows 2000','Internet Explorer 5.5'),(433,'Website Quester - www.asona.org',NULL,NULL),(434,'Mozilla/4.0 (compatible; MSIE 5.01; Windows 95; {World Online})','Windows 98','Internet Explorer 5.0'),(435,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Win 9x 4.90; Creative)','Windows 98','Internet Explorer 5.5'),(436,'Mozilla/4.0 (compatible; MSIE 5.01; Windows 95; MSOCD; AtHomeNL191)','Windows 98','Internet Explorer 5.0'),(437,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Win 9x 4.90; Het Net 3.0)','Windows 98','Internet Explorer 5.5'),(438,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; MSN 6.1; MSNbMSFT; MSNmnl-nl; MSNc0z)','Windows 98','Internet Explorer 5.5'),(439,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; MSN 6.1; MSNbMSFT; MSNmnl-nl; MSNc00)','Windows 98','Internet Explorer 5.5'),(440,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; WanadooNL40)','Windows 98','Internet Explorer 6.0'),(441,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.0; c\'t)','Windows 2000','Internet Explorer 6.0'),(442,'Mozilla/4.7 [en]C-CCK-MCD NSCPCD47  (Win95; I)','Windows 95','Netscape 4.7'),(443,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; CHWIE_NL60)','Windows XP','Internet Explorer 6.0'),(444,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 95; Wegener)','Windows 98','Internet Explorer 5.5'),(445,'Mozilla/4.0 (compatible; MSIE 5.01; Windows NT 5.0; Q312461)','Windows 2000','Internet Explorer 5.0'),(446,'Mozilla/4.61 [en] (Win98; I)','Windows 98','Netscape 4.61'),(447,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; AtHomeNL191; T312461)','Windows 98','Internet Explorer 5.5'),(448,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; Win 9x 4.90; {VIZZAVI0101})','Windows 98','Internet Explorer 6.0'),(449,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; MSN 6.1; MSNbMSFT; MSNmnl-nl; MSNc0z)','Windows XP','Internet Explorer 6.0'),(450,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; MSN 6.1; MSNbMSFT; MSNmnl-nl; MSNc00)','Windows XP','Internet Explorer 6.0'),(451,'Mozilla/4.51 [de]C-CCK-MCD DT  (Win95; I)','Windows 95','Netscape 4.51'),(452,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; StudentenNet)','Windows 98','Internet Explorer 5.5'),(453,'Mozilla/4.0 (compatible; MSIE 5.0; MSN 2.5; Windows 98; DigExt)','Windows 98','Internet Explorer 5.0'),(454,'Mozilla/4.7 [de] (WinNT; I)','Windows NT','Netscape 4.7'),(455,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; Win 9x 4.90; {World Online})','Windows 98','Internet Explorer 6.0'),(456,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; WISH; DigExt; ZON; {World Online}; NOKNOK)','Windows 98','Internet Explorer 5.0'),(457,'Wget/1.5.3',NULL,'Wget 1.5'),(458,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Win 9x 4.90; MSIECrawler)','Windows 98','Internet Explorer 5.5'),(459,'ASPSeek/1.2.5',NULL,NULL),(460,'Mozilla/4.0 (compatible; MSIE 6.0; CS 2000; Windows NT 5.1; {VIZZAVI0101})','Windows XP','Internet Explorer 6.0'),(461,'Mozilla/4.7 [nl] (Win98; U)','Windows 98','Netscape 4.7'),(462,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; DigExt; {Vizzavi052001})','Windows 98','Internet Explorer 5.0'),(463,'Mozilla/4.0 (compatible; MSIE 5.01; Windows 95; NCRVNET)','Windows 98','Internet Explorer 5.0'),(464,'Mozilla/5.0 (Macintosh; U; PPC; en-US; rv:0.9.4) Gecko/20011022 Netscape6/6.2','Mac OS','Netscape 6.2'),(465,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 95; HvA)','Windows 95','Internet Explorer 5.5'),(466,'Mozilla/4.0 (compatible; MSIE 4.0; Windows 95)','Windows 98','Internet Explorer 4.0'),(467,'Mozilla/4.61 [en] (WinNT; I)','Windows NT','Netscape 4.61'),(468,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; Hotbar 2.0)','Windows 98','Internet Explorer 6.0'),(469,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.0; Q312461; c\'t)','Windows 2000','Internet Explorer 6.0'),(470,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.0; c\'t; Q312461)','Windows 2000','Internet Explorer 6.0'),(471,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; WISH; DigExt; MSNCOM)','Windows 98','Internet Explorer 5.0'),(472,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Win 9x 4.90; {World Online})','Windows 98','Internet Explorer 5.5'),(473,'Mozilla/4.0 (compatible; MSIE 5.01; Windows 98; Wanadoo NL; KITV4 Wanadoo)','Windows 98','Internet Explorer 5.0'),(474,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; ZeelandNet TeleFree V.1; DigExt)','Windows 98','Internet Explorer 5.0'),(475,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; MSOCD; AtHomeNL191; Hotbar 3.0)','Windows 98','Internet Explorer 6.0'),(476,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Wanadoo 5.1; Wanadoo 5.3)','Windows 98','Internet Explorer 5.5'),(477,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; Hotbar 3.0; Q312461)','Windows 98','Internet Explorer 6.0'),(478,'Mozilla/4.0 (compatible; MSIE 5.0; Windows NT; DigExt; MSNCOM)','Windows NT','Internet Explorer 5.0'),(479,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; MSNCOM)','Windows 98','Internet Explorer 6.0'),(480,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; MSOCD)','Windows 98','Internet Explorer 5.5'),(481,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; MSOCD)','Windows XP','Internet Explorer 6.0'),(482,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; WanadooNL31)','Windows 98','Internet Explorer 5.5'),(483,'Mozilla/4.0 (compatible; MSIE 5.01; Windows NT; EK501002)','Windows NT','Internet Explorer 5.0'),(484,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; Planet-ADSL)','Windows 98','Internet Explorer 6.0'),(485,'Mozilla/4.6 (Macintosh; I; PPC)','Mac OS','Netscape 4.6'),(486,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.0; MSOCD; AtHomeNL191)','Windows 2000','Internet Explorer 6.0'),(487,'Mozilla/4.0 (compatible; MSIE 5.01; Windows NT; EK501002; EK60)','Windows NT','Internet Explorer 5.0'),(488,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; KITXP40NL)','Windows XP','Internet Explorer 6.0'),(489,'AliveProxy',NULL,NULL),(490,'Mozilla/4.0 (compatible; MSIE 5.5b1; Mac_PowerPC)','Mac OS','Internet Explorer 5.5'),(491,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Win 9x 4.90; WanadooEN31)','Windows 98','Internet Explorer 5.5'),(492,'Scooter-ARS-1.1',NULL,NULL),(493,'Mozilla/4.6 [en] (Win95; I)','Windows 95','Netscape 4.6'),(494,'Mozilla/4.0 (compatible; MSIE 5.5; Windows NT 4.0; Hotbar 3.0)','Windows NT 4.0','Internet Explorer 5.5'),(495,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; ZON; Q312461)','Windows 98','Internet Explorer 6.0'),(496,'Mozilla/4.0 (compatible; MSIE 5.0; CS 2000; Windows 95; DigExt)','Windows 98','Internet Explorer 5.0'),(497,'Mozilla/4.0 (compatible; MSIE 5.01; Windows 95; DigExt)','Windows 98','Internet Explorer 5.0'),(498,'Mozilla/6.0',NULL,NULL),(499,'Mozilla/4.61 [en] (OS/2; U)','OS/2','Netscape 4.61'),(500,'Mozilla/4.51 [es] (Win98; I)','Windows 98','Netscape 4.51'),(501,'Mozilla/4.0 (compatible; MSIE 5.0b1; Mac_PowerPC)','Mac OS','Internet Explorer 5.0'),(502,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 95; DigExt; {World Online})','Windows 98','Internet Explorer 5.0'),(503,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; DigExt; Wanadoo NL; KITV4 Wanadoo)','Windows 98','Internet Explorer 5.0'),(504,'Mozilla/4.0 (compatible; MSIE 5.0; Windows NT 4.0) Opera 6.01  [en]','Windows NT 4.0','Internet Explorer 5.0'),(505,'Mozilla/4.0 (compatible; MSIE 5.0; AOL 5.0; Windows 98; DigExt)','Windows 98','Internet Explorer 5.0'),(506,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; DigExt; WISH; ZON)','Windows 98','Internet Explorer 5.0'),(507,'LWP::Simple/5.53',NULL,NULL),(508,'Spida/0.1',NULL,NULL),(509,'Mozilla/4.0 (compatible; MSIE 6.0; CS 2000; Windows 98; H010818)','Windows 98','Internet Explorer 6.0'),(510,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Win 9x 4.90; Zon CD-rom versie 2)','Windows 98','Internet Explorer 5.5'),(511,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; DigExt; Planet Internet)','Windows 98','Internet Explorer 5.0'),(512,'Mozilla/4.0 (compatible; MSIE 4.5; Mac_PowerPC; e504207PDI_IF_2K)','Mac OS','Internet Explorer 4.5'),(513,'libwww-perl/5.51',NULL,NULL),(514,'Mozilla/4.0 (compatible; MSIE ; Mac_PPC)','Mac OS',NULL),(515,'Opera/6.01 (Windows ME; U)  [ja]','Windows ME','Opera 6.0'),(516,'Mozilla/4.7 [en] (WinNT; U)','Windows NT','Netscape 4.7'),(517,'Mozilla/5.0 (Macintosh; U; PPC Mac OS X; en-US; rv:0.9.4) Gecko/20011022 Netscape6/6.2','Mac OS X','Netscape 6.2'),(518,'Mozilla/4.0 (compatible; MSIE 5.5; Windows NT 5.0; DigExt; DT)','Windows 2000','Internet Explorer 5.5'),(519,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98) Opera 6.0  [en]','Windows 98','Internet Explorer 5.0'),(520,'Mozilla/4.0 (compatible; MSIE 5.01; Windows NT 5.0; HCCnet)','Windows 2000','Internet Explorer 5.0'),(521,'Mozilla/4.7 [en] (Win98; I)','Windows 98','Netscape 4.7'),(522,'Mozilla/4.0 (compatible; MSIE 5.5; Windows NT 5.0; MSNCOM)','Windows 2000','Internet Explorer 5.5'),(523,'Mozilla/4.5 (Macintosh; I; PPC)','Mac OS','Netscape 4.5'),(524,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Concepts ICT Excellence)','Windows 98','Internet Explorer 5.5'),(525,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Win 9x 4.90; ZON)','Windows 98','Internet Explorer 5.5'),(526,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 95; FREELER1.1; DigExt)','Windows 98','Internet Explorer 5.0'),(527,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; FreeSurf myweb.nl v1.0.1; T312461)','Windows 98','Internet Explorer 5.5'),(528,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; DigExt; Het Net 2.0)','Windows 98','Internet Explorer 5.0'),(529,'Bilbo/2.3a-UNIX','Unix',NULL),(530,'Mozilla/4.73 (Macintosh; U; PPC)','Mac OS','Netscape 4.73'),(531,'FAST-WebCrawler/3.3 (crawler@fast.no; http://fast.no/support.php?c=faqs/crawler)',NULL,NULL),(532,'Mozilla/4.0 (0000000000; 0000 0000; 00000000000)',NULL,'Netscape 4.0'),(533,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; Q312461; NOKNOK)','Windows XP','Internet Explorer 6.0'),(534,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; PKBL008; Hi Way, powered by Planet Internet; DigExt)','Windows 98','Internet Explorer 5.0'),(535,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; KITV4 Wanadoo; DigExt)','Windows 98','Internet Explorer 5.0'),(536,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 4.0; Q312461)','Windows NT 4.0','Internet Explorer 6.0'),(537,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98) WebWasher 3.0','Windows 98','Internet Explorer 6.0'),(538,'Mozilla/4.5 [fr] (Macintosh; I; PPC)','Mac OS','Netscape 4.5'),(539,'Mozilla/2.0 (compatible; MS FrontPage 4.0)',NULL,NULL),(540,'MSFrontPage/4.0',NULL,NULL),(541,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; Win 9x 4.90; Het Net 4.0)','Windows 98','Internet Explorer 6.0'),(542,'Mozilla/4.0 (compatible; MSIE 5.5; Windows NT 4.0; Tucows)','Windows NT 4.0','Internet Explorer 5.5'),(543,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; Q312461; Hotbar 3.0)','Windows XP','Internet Explorer 6.0'),(544,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; TNET5.0NL)','Windows 98','Internet Explorer 5.5'),(545,'Mozilla/4.0 (compatible; MSIE 5.01; CS 2000; Windows 95; {World Online})','Windows 98','Internet Explorer 5.0'),(546,'Mozilla/4.04 [en] (WinNT; I)','Windows NT','Netscape 4.04'),(547,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; ZON)','Windows 98','Internet Explorer 5.5'),(548,'Mozilla/4.0 (compatible; MSIE 5.5; Windows NT 5.0; Hotbar 3.0)','Windows 2000','Internet Explorer 5.5'),(549,'Mozilla/4.5 [en]C-CCK-MCD {World Online}  (Win95; I)','Windows 95','Netscape 4.5'),(550,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; TNET5.0FR)','Windows 98','Internet Explorer 5.5'),(551,'Mozilla/4.61 [en] (Win95; I)','Windows 95','Netscape 4.61'),(552,'Mozilla/4.06 [en] (Win98; I)','Windows 98','Netscape 4.06'),(553,'Mozilla/5.0 (Macintosh; U; PPC; en-US; rv:0.9.9) Gecko/20020311','Mac OS','Mozilla 0.9'),(554,'LinkLint-checkonly/2.3.4',NULL,NULL),(555,'Mozilla/5.0 (compatible; Konqueror/2.2.1; Linux)','Linux','Konqueror 2.2.1'),(556,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; AtHomeNL191)','Windows XP','Internet Explorer 6.0'),(557,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; DigExt; {World Online}; FreeSurf tmfweb.nl v1.0.1)','Windows 98','Internet Explorer 5.0'),(558,'Mozilla/4.0 (compatible; MSIE 5.01; Windows NT; B.V. Ontwikkelingsmaatschappij C.C.M.; B.V. Ontwikkelingsmij. C.C.M.)','Windows NT','Internet Explorer 5.0'),(559,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; Hotbar 2.0; WanadooNL40)','Windows 98','Internet Explorer 6.0'),(560,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; ZDNetIndia.com browser)','Windows 98','Internet Explorer 5.5'),(561,'Mozilla/4.0 (compatible; MSIE 5.5; Windows NT)','Windows NT','Internet Explorer 5.5'),(562,'Mozilla/3.0 (Vagabondo/1.0 MT; webagent@wise-guys.nl; http://webagent.wise-guys.nl/)',NULL,NULL),(563,'Mozilla/4.07 [en] (Win98; I)','Windows 98','Netscape 4.07'),(564,'Mozilla/4.0 (compatible; MSIE 4.01; Windows 98; HCCnet Internet Explorer)','Windows 98','Internet Explorer 4.01'),(565,'Mozilla/4.0 (compatible; MSIE 5.0; CS 2000; Windows 98; DigExt; ZON)','Windows 98','Internet Explorer 5.0'),(566,'Ideare - SignSite/1.2',NULL,NULL),(567,'Wget/1.7',NULL,'Wget 1.7'),(568,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 95; DigExt; FreeSurf myweb.nl v1.0.1)','Windows 98','Internet Explorer 5.0'),(569,'Mozilla/4.0 (compatible; MSIE 4.01; Windows 95; EUnet)','Windows 98','Internet Explorer 4.01'),(570,'Mozilla/4.7 [nl] (WinNT; I)','Windows NT','Netscape 4.7'),(571,'Mozilla/4.5 [nl] (Win98; I)','Windows 98','Netscape 4.5'),(572,'Mozilla/4.01 [en] (Win95; I)','Windows 95','Netscape 4.01'),(573,'Mozilla/4.5 [en] (WinNT; I) via Avirt Gateway Server v4.2','Windows NT','Netscape 4.5'),(574,'Mozilla/4.0 (compatible; MSIE 5.0; www.galaxy.com; www.adultalliances.com)',NULL,'Internet Explorer 5.0'),(575,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; Wanadoo cable; H010818)','Windows 98','Internet Explorer 6.0'),(576,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; {World Online}; Q312461)','Windows XP','Internet Explorer 6.0'),(577,'larbin_2.6.1 larbin2.6.1@unspecified.mail',NULL,NULL),(578,'larbin_2.6.1 (larbin2.6.1@unspecified.mail)',NULL,NULL),(579,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; WorldOnline-nl-v)','Windows 98','Internet Explorer 6.0'),(580,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Win 9x 4.90; Cable Wanadoo)','Windows 98','Internet Explorer 5.5'),(581,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; brip1)','Windows 98','Internet Explorer 5.5'),(582,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; DigExt; WISH)','Windows 98','Internet Explorer 5.0'),(583,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; Win 9x 4.90; Het Net 3.2)','Windows 98','Internet Explorer 6.0'),(584,'Mozilla/4.5 [en] (Win98; I)','Windows 98','Netscape 4.5'),(585,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 95)','Windows 98','Internet Explorer 5.0'),(586,'Mozilla/4.75 [en] (Windows 5.0; U)','Windows 2000','Netscape 4.75'),(587,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; Q312461; ZON; DVD Owner)','Windows XP','Internet Explorer 6.0'),(588,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; Hi Way, powered by Planet Internet; Planet Internet; ZON; Het Net 3.0; DigExt)','Windows 98','Internet Explorer 5.0'),(589,'Mozilla/5.0 (Windows; U; Win98; en-GB; rv:0.9.2) Gecko/20010726 Netscape6/6.1','Windows 98','Netscape 6.1'),(590,'Mozilla/3.01C-WXS-32  (Win95; I)','Windows 95','Netscape 3.01'),(591,'Mozilla/4.0 (compatible; MSIE 4.01; Windows 98; {World Online})','Windows 98','Internet Explorer 4.01'),(592,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Win 9x 4.90; WanadooNL31; T312461)','Windows 98','Internet Explorer 5.5'),(593,'Mozilla/4.51 [nl] (WinNT; I)','Windows NT','Netscape 4.51'),(594,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Win 9x 4.90; T312461)','Windows 98','Internet Explorer 5.5'),(595,'Mozilla/4.0 (compatible; MSIE 5.01; Windows NT 5.0; (R1 1.1))','Windows 2000','Internet Explorer 5.0'),(596,'multiBlocker browser',NULL,NULL),(597,'Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US; rv:0.9.9) Gecko/20020311','Windows XP','Mozilla 0.9'),(598,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; Q312461; Wanadoo cable)','Windows 98','Internet Explorer 6.0'),(599,'Mozilla/4.0 (compatible; MSIE 5.5; Windows NT 5.0; Creative)','Windows 2000','Internet Explorer 5.5'),(600,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; {World Online}; T312461)','Windows 98','Internet Explorer 5.5'),(601,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; MSOCD; AtHomeNL191; Q312461)','Windows 98','Internet Explorer 6.0'),(602,'Mozilla/4.03 [nl] (Win95; I)','Windows 95','Netscape 4.03'),(603,'Mozilla/4.0 (compatible; MSIE 5.5; Windows NT 4.0; IE55SP2 (19092001))','Windows NT 4.0','Internet Explorer 5.5'),(604,'Lynx/2.8.4rel.1 libwww-FM/2.14 SSL-MM/1.4.1 OpenSSL/0.9.6a','','Lynx 2.8'),(605,'Mozilla/4.0 (compatible; MSIE 5.01; Windows NT 5.0; FreeSurf iweb.nl v1.0.2)','Windows 2000','Internet Explorer 5.0'),(606,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; DigExt; Hotbar 2.0)','Windows 98','Internet Explorer 5.0'),(607,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 95) Opera 6.01  [nl]','Windows 98','Internet Explorer 5.0'),(608,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; DigExt; FreeSurf tmfweb.nl v1.0.2)','Windows 98','Internet Explorer 5.0'),(609,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; Win 9x 4.90; Hotbar 3.0; ZON)','Windows 98','Internet Explorer 6.0'),(610,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 95; T312461; LUMC)','Windows 98','Internet Explorer 5.5'),(611,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 4.0; Q312461; (R1 1.1))','Windows NT 4.0','Internet Explorer 6.0'),(612,'Mozilla/4.73 [en] (Win98; U)','Windows 98','Netscape 4.73'),(613,'Mozilla/4.0 (compatible; MSIE 5.01; Windows 98; Compaq)','Windows 98','Internet Explorer 5.0'),(614,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Wanadoo NL)','Windows 98','Internet Explorer 5.5'),(615,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; DigExt; AltaVista 1.01.01)','Windows 98','Internet Explorer 5.0'),(616,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 95; Het Net 3.2)','Windows 98','Internet Explorer 5.5'),(617,'Mozilla/4.0 (compatible; MSIE 5.0; CS 2000; Windows 98; DigExt; FreeSurf myweb.nl v1.0.1)','Windows 98','Internet Explorer 5.0'),(618,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 95) VoilaBot; 1.6','Windows 98','Internet Explorer 5.0'),(619,'Mozilla/4.0 (compatible; MSIE 4.01; Windows 95; HCCnet Internet Explorer)','Windows 98','Internet Explorer 4.01'),(620,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; DT)','Windows 98','Internet Explorer 5.5'),(621,'Mozilla/4.0 (compatible; MSIE 5.01; Windows 98; WanadooNL; KITV4 Wanadoo)','Windows 98','Internet Explorer 5.0'),(622,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; Win 9x 4.90; FREELER1.1)','Windows 98','Internet Explorer 6.0'),(623,'Mozilla/4.0 (compatible; MSIE 6.0; CS 2000; Windows 98; Win 9x 4.90; Q312461)','Windows 98','Internet Explorer 6.0'),(624,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; DigExt; FREELER1)','Windows 98','Internet Explorer 5.0'),(625,'Mozilla/5.0 (Windows; U; WinNT4.0; en-US; rv:0.9.4) Gecko/20011128 Netscape6/6.2.1','Windows NT 4.0','Netscape 6.2'),(626,'Mozilla/4.7 [ja] (Macintosh; I; PPC)','Mac OS','Netscape 4.7'),(627,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 95; MSN 6.1; MSNbMSFT; MSNmnl-nl; MSNcOTH)','Windows 98','Internet Explorer 5.5'),(628,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; {Vizzavi052001})','Windows 98','Internet Explorer 5.5'),(629,'Mozilla/4.77 [en] (X11; U; Linux 2.2.19 i686)','Linux','Netscape 4.77'),(630,'SearchTone2.0 - IDEARE',NULL,NULL),(631,'Mozilla/4.0 (compatible; MSIE 5.50; Windows NT; SiteKiosk 4.9; SiteCoach 1.0)','Windows NT','Internet Explorer 5.5'),(632,'Mozilla/4.77 [en]C-CCK-MCD HSA  (WinNT; U)','Windows NT','Netscape 4.77'),(633,'Mozilla/5.0 (Windows; U; Win98; en-US; rv:0.9.4) Gecko/20011019 Netscape6/6.2','Windows 98','Netscape 6.2'),(634,'Mozilla/4.0 (compatible; MSIE 5.5; Windows NT 5.0; T312461; NetCaptor 6.5.0)','Windows 2000','Internet Explorer 5.5'),(635,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; DigExt; Open Universiteit Nederland; installatie CD-rom)','Windows 98','Internet Explorer 5.0'),(636,'Mozilla/4.0 (compatible; MSIE 5.5; Windows NT 5.0; NetCaptor 6.5.0B7)','Windows 2000','Internet Explorer 5.5'),(637,'Mozilla/4.0 (compatible; MSIE 5.01; Windows NT 5.0; MSIECrawler)','Windows 2000','Internet Explorer 5.0'),(638,'Mozilla/4.0 (compatible; MSIE 5.0; Windows NT; MSNDE)','Windows NT','Internet Explorer 5.0'),(639,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; StarBand Version 1.0)','Windows 98','Internet Explorer 5.5'),(640,'Mozilla/4.0 (compatible; MSIE 5.01; Windows 98; KSC)','Windows 98','Internet Explorer 5.0'),(641,'Mozilla/4.6 (Macintosh; U; PPC)','Mac OS','Netscape 4.6'),(642,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; Win 9x 4.90; yie6)','Windows 98','Internet Explorer 6.0'),(643,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; NOKNOK)','Windows 98','Internet Explorer 5.5'),(644,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Win 9x 4.90; Parool INTERNET)','Windows 98','Internet Explorer 5.5'),(645,'Mozilla/4.0 (compatible; MSIE 5.0; Mac_PowerPC; e504460WanadooNL)','Mac OS','Internet Explorer 5.0'),(646,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; DigExt; AtHomeNL0107)','Windows 98','Internet Explorer 5.0'),(647,'Mozilla/4.04 [nl] (Win95; I ;Nav)','Windows 95','Netscape 4.04'),(648,'Mozilla/4.0 (compatible; MSIE 5.5; Windows NT 4.0; MSIECrawler)','Windows NT 4.0','Internet Explorer 5.5'),(649,'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:0.9.8) Gecko/20020204','Linux','Mozilla 0.9'),(650,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; Charter B1)','Windows 98','Internet Explorer 6.0'),(651,'Mozilla/4.08 [en] (Win98; U ;Nav)','Windows 98','Netscape 4.08'),(652,'Mozilla/4.03 [de] (Macintosh; I; PPC)','Mac OS','Netscape 4.03'),(653,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; AtHomeNL191)','Windows 98','Internet Explorer 5.5'),(654,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; NRCnet Internet)','Windows 98','Internet Explorer 5.5'),(655,'Mozilla/4.0 (compatible; MSIE 5.01; Windows 95; Demon NL v9810.01)','Windows 98','Internet Explorer 5.0'),(656,'Mozilla/4.0 (compatible; MSIE 5.0; www.galaxy.com; www.makeblijde.nl)',NULL,'Internet Explorer 5.0'),(657,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; Hotbar 3.0; MSN 6.1; MSNbMSFT; MSNmnl-nl; MSNcOTH)','Windows XP','Internet Explorer 6.0'),(658,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; MSN 6.1; MSNbMSFT; MSNmnl-nl; MSNcOTH)','Windows XP','Internet Explorer 6.0'),(659,'Mozilla/4.0 (compatible; MSIE 5.5; AOL 7.0; Windows 98)','Windows 98','Internet Explorer 5.5'),(660,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; DigExt; u1user)','Windows 98','Internet Explorer 5.0'),(661,'Mozilla/4.7 [en-gb] (Win98; U)','Windows 98','Netscape 4.7'),(662,'LWP::Simple/5.48',NULL,NULL),(663,'gazz/2.1 (gazz@nttrd.com)',NULL,NULL),(664,'Mozilla/4.0 (compatible; MSIE 4.01; Windows 98; Het Net 2.0)','Windows 98','Internet Explorer 4.01'),(665,'Mozilla/4.7 (compatible; Whizbang)',NULL,NULL),(666,'My_Agent',NULL,NULL),(667,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.0; AtHomeNL191; Q312461; Hotbar 3.0)','Windows 2000','Internet Explorer 6.0'),(668,'Mozilla/4.07 [en] (X11; I; Linux 2.2.3 i686)','Linux','Netscape 4.07'),(669,'Mozilla/4.0 (compatible; MSIE 5.01; Windows 98; {World Online - Abn Amro})','Windows 98','Internet Explorer 5.0'),(670,'Mozilla/4.04 [en] (Win95; I ;Nav)','Windows 95','Netscape 4.04'),(671,'Mozilla/2.0 (compatible; MSIE 3.01; CS 2000; Windows 95)','Windows 98','Internet Explorer 3.01'),(672,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; {VIZZAVI0101}; KITV4 Wanadoo)','Windows 98','Internet Explorer 6.0'),(673,'Mozilla/5.0 (Windows; U; WinNT4.0; en-US; m18) Gecko/20010131 Netscape6/6.01','Windows NT 4.0','Netscape 6.01'),(674,'Mozilla/4.75 [en] (Windows NT 5.0; U)','Windows 2000','Netscape 4.75'),(675,'Mozilla/4.0 (compatible; MSIE 5.01; Windows NT 5.0; DigExt)','Windows 2000','Internet Explorer 5.0'),(676,'Mozilla/4.77 [en] (Win95; U)','Windows 95','Netscape 4.77'),(677,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; H010818)','Windows 98','Internet Explorer 6.0'),(678,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Win 9x 4.90; WanadooNL31)','Windows 98','Internet Explorer 5.5'),(679,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 95) TrueRobot; 1.5','Windows 98','Internet Explorer 5.0'),(680,'Mozilla/4.0_(compatible;_MSIE_5.0;_Windows_95)_TrueRobot/1.4 libwww/5.2.8','Windows 95','Internet Explorer 5.0'),(681,'Doerak/1.0libwww-perl/5.49',NULL,NULL),(682,'Internet Explorer 4.01',NULL,'Internet Explorer 4.0'),(683,'Mozilla/4.03 [fr] (Win95; I)','Windows 95','Netscape 4.03'),(684,'Mozilla/4.05 (Macintosh; I; PPC, Nav)','Mac OS','Netscape 4.05'),(685,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; NRCnet Internet; Q312461)','Windows 98','Internet Explorer 6.0'),(686,'Mozilla/4.76 [en] (Win95; U)','Windows 95','Netscape 4.76'),(687,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.0; ZON; Q312461)','Windows 2000','Internet Explorer 6.0'),(688,'Mozilla/4.7 [nl]C-CCK-MCD   (Win98; U)','Windows 98','Netscape 4.7'),(689,'Mozilla/4.0 (compatible; MSIE 5.0; Windows NT; DigExt; WMN)','Windows NT','Internet Explorer 5.0'),(690,'appie 1.1 (www.walhello.com)',NULL,NULL),(691,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; AtHomeI0107)','Windows 98','Internet Explorer 5.5'),(692,'Mozilla/4.0 (compatible; MSIE 5.01; Windows NT; Fortis Bank)','Windows NT','Internet Explorer 5.0'),(693,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 95; CWIE55NL60; CHWIE_NL60)','Windows 98','Internet Explorer 5.5'),(694,'Mozilla/4.73 [en] (WinNT; I)','Windows NT','Netscape 4.73'),(695,'Mozilla/4.72 [en] (Win95; I)','Windows 95','Netscape 4.72'),(696,'Mozilla/4.0 (compatible; MSIE 5.0; CS 2000; Windows 98; DigExt; Planet Internet PIM 2)','Windows 98','Internet Explorer 5.0'),(697,'Mozilla/4.0 (compatible; MSIE 5.01; Windows 98; FREELER1.1)','Windows 98','Internet Explorer 5.0'),(698,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Het Net 4.0; T312461)','Windows 98','Internet Explorer 5.5'),(699,'Mozilla/4.0 (compatible; MSIE 5.01; Windows 98; PKBL008)','Windows 98','Internet Explorer 5.0'),(700,'Mozilla/4.0 (compatible; SuperCleaner 2.32; Windows NT 5.1)','Windows XP',NULL),(701,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; Creative)','Windows 98','Internet Explorer 6.0'),(702,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Win 9x 4.90; CHWIE_NL60)','Windows 98','Internet Explorer 5.5'),(703,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Win 9x 4.90; SKY12)','Windows 98','Internet Explorer 5.5'),(704,'EPOC32-WTL/2.2 Crystal/6.0 STNC-WTL/6.0(611)',NULL,NULL),(705,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; {VIZZAVI0101})','Windows 98','Internet Explorer 6.0'),(706,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; WanadooKITXP40NL)','Windows XP','Internet Explorer 6.0'),(707,'Mozilla/5.0 (Windows; U; Win98; en-US; rv:0.9.4) Gecko/20011019 Netscape6/6.2 (CK-PCUser)','Windows 98','Netscape 6.2'),(708,'Mozilla/4.0 (compatible; MSIE 5.01; Windows 95; KITV4 Wanadoo)','Windows 98','Internet Explorer 5.0'),(709,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 95; Het Net 4.0)','Windows 98','Internet Explorer 5.5'),(710,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; 24hours01)','Windows 98','Internet Explorer 5.5'),(711,'Mozilla/4.0 (compatible; MSIE 5.01; Windows 95; AtHomeNL191)','Windows 98','Internet Explorer 5.0'),(712,'Lynx/2.8.4rel.1 libwww-FM/2.14 SSL-MM/1.4.1 OpenSSL/0.9.6b',NULL,'Lynx 2.8'),(713,'Mozilla/4.0 (compatible; MSIE 5.01; Windows NT; gemland)','Windows NT','Internet Explorer 5.0'),(714,'Scooter-3.2.DIL',NULL,NULL),(715,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; Hi Way, powered by Planet Internet; DigExt)','Windows 98','Internet Explorer 5.0'),(716,'Mozilla/5.0 (Macintosh; U; PPC; en-US; rv:0.9.2) Gecko/20010726 Netscape6/6.1','Mac OS','Netscape 6.1'),(717,'sitechecker',NULL,NULL),(718,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; AtHomeNL191; Hotbar 3.0)','Windows 98','Internet Explorer 6.0'),(719,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; Q312461; TUCOWS)','Windows 98','Internet Explorer 5.0'),(720,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; TNET5.0NL)','Windows 98','Internet Explorer 6.0'),(721,'Mozilla/4.08 [en] (WinNT; I ;Nav)','Windows NT','Netscape 4.08'),(722,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; PKBL008; ZON)','Windows 98','Internet Explorer 5.5'),(723,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; FREELER1.2)','Windows 98','Internet Explorer 5.5'),(724,'Mozilla/5.0 (Windows; U; WinNT4.0; en-US; rv:1.0rc1) Gecko/20020417','Windows NT 4.0','Mozilla 1.0'),(725,'Mozilla/4.0 (compatible; MSIE 6.0b; Windows NT 5.0; .NET CLR 1.0.2914)','Windows 2000','Internet Explorer 6.0'),(726,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98) Opera 6.03  [en]','Windows 98','Internet Explorer 5.0'),(727,'ProxyHunter',NULL,NULL),(728,'Mozilla/5.0 (Windows; U; Win98; de-DE; rv:0.9.4) Gecko/20011128 Netscape6/6.2.1','Windows 98','Netscape 6.2'),(729,'Mozilla/5.0 (Windows; U; Win98; en-GB; rv:0.9.4) Gecko/20011019 Netscape6/6.2','Windows 98','Netscape 6.2'),(730,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Zon CD-rom versie 2)','Windows 98','Internet Explorer 5.5'),(731,'Mozilla/4.78 [en]C-CCK-MCD {World Online}  (Win95; U)','Windows 95','Netscape 4.78'),(732,'Mozilla/4.61 [en] (X11; I; SunOS 5.6 sun4u; Nav)','SunOS','Netscape 4.61'),(733,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; Cablewanadoo premium and plus)','Windows XP','Internet Explorer 6.0'),(734,'Mozilla/4.0 (compatible; MSIE 5.01; Windows 98; Hotbar 3.0)','Windows 98','Internet Explorer 5.0'),(735,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; TNET5.0NL)','Windows XP','Internet Explorer 6.0'),(736,'Mozilla/4.0 (compatible; MSIE 5.5; Windows NT 5.0) Active Cache Request','Windows 2000','Internet Explorer 5.5'),(737,'Mozilla/4.0 (compatible; MSIE 5.5; Windows NT 5.0; Zon CD-rom versie 2)','Windows 2000','Internet Explorer 5.5'),(738,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; DigExt; ZON; KITV4 Wanadoo)','Windows 98','Internet Explorer 5.0'),(739,'Mozilla/4.0 (compatible; MSIE 5.5; Windows NT 5.0; NOKNOK)','Windows 2000','Internet Explorer 5.5'),(740,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; Zon CD-rom versie 2)','Windows XP','Internet Explorer 6.0'),(741,'Mozilla/4.0 (compatible; MSIE 5.0; Windows NT; Girafabot; girafabot at girafa dot com; http://www.girafa.com)','Windows NT','Internet Explorer 5.0'),(742,'Mozilla/4.0 (compatible; MSIE 5.0; Windows NT; DigExt; Mozilla/4.0 (compatible; MSIE 5.0; Win32))','Windows NT','Internet Explorer 5.0'),(743,'Mozilla/4.79 [en] (X11; U; Linux 2.2.19 i686; Nav)','Linux','Netscape 4.79'),(744,'Mozilla/5.0 (Windows; U; Win 9x 4.90; en-US; rv:1.0rc2) Gecko/20020512 Netscape/7.0b1','Windows ME','Netscape 7.0b1'),(745,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; T312461)','Windows 98','Internet Explorer 6.0'),(746,'Mozilla/4.0 (compatible; MSIE 5.5; Windows NT 5.0; H010818; T312461; Hewlett-Packard IE5.5-SP2)','Windows 2000','Internet Explorer 5.5'),(747,'Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US; rv:1.0rc2) Gecko/20020512 Netscape/7.0b1','Windows XP','Netscape 7.0'),(748,'Mozilla/4.5 [es] (Win95; I)','Windows 95','Netscape 4.5'),(749,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; T312461; Crazy Browser 1.0.5)','Windows 98','Internet Explorer 6.0'),(750,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Win 9x 4.90; NOKNOK)','Windows 98','Internet Explorer 5.5'),(751,'Mozilla/5.0 (Windows; U; Win 9x 4.90; en-US; m18) Gecko/20010131 Netscape6/6.01','Windows ME','Netscape 6.01'),(752,'Mozilla/4.0 (compatible; MSIE 5.01; Windows 98; Planet Internet PIM 2)','Windows 98','Internet Explorer 5.0'),(753,'Mozilla/5.0 (Windows; U; Windows NT 5.0; en-US; rv:0.9.4.1) Gecko/20020508 Netscape6/6.2.3','Windows 2000','Netscape 6.2'),(754,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 95; Zon CD-rom versie 2)','Windows 98','Internet Explorer 5.5'),(755,'Mozilla/4.0 (compatible; MSIE 5.01; Windows NT 5.0; HCCnet Internet Explorer)','Windows 2000','Internet Explorer 5.0'),(756,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; Win 9x 4.90; Het Net 3.2; Q312461)','Windows 98','Internet Explorer 6.0'),(757,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; AtHomeNL0107)','Windows XP','Internet Explorer 6.0'),(758,'Mozilla/4.73 [en] (Win98; I)','Windows 98','Netscape 4.73'),(759,'Mozilla/4.79 [en] (WinNT; U)','Windows NT','Netscape 4.79'),(760,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Win 9x 4.90; Wanadoo cable; H010818; Hotbar 3.0)','Windows 98','Internet Explorer 5.5'),(761,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; surfEU DE S3)','Windows 98','Internet Explorer 5.5'),(762,'Mozilla/4.0 (compatible; MSIE 6.0; AOL 7.0; Windows 98)','Windows 98','Internet Explorer 6.0'),(763,'FAST-WebCrawler/3.6 (atw-crawler at fast dot no; http://fast.no/support/crawler.asp)',NULL,'FAST-WebCrawler 3.6'),(764,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; Planet Internet PIM 2)','Windows XP','Internet Explorer 6.0'),(765,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; PKBL008)','Windows 98','Internet Explorer 6.0'),(766,'Scooter-3.2.QA',NULL,NULL),(767,'Mozilla/5.0 (Windows; U; Win98; en-US; rv:1.0rc2) Gecko/20020512 Netscape/7.0b1','Windows 98','Netscape 7.0'),(768,'Mozilla/4.0 (compatible; MSIE 5.0; Windows NT; DigExt; DTS Agent','Windows NT 4.0','Internet Explorer 5.0'),(769,'Mozilla/4.5 [en] (WinNT; I)','Windows NT','Netscape 4.5'),(770,'Mozilla/4.0 (compatible; MSIE 5.01; Windows NT; {World Online})','Windows NT','Internet Explorer 5.0'),(771,'Mozilla/4.0 (compatible; MSIE 5.01; Windows NT; MinEZ)','Windows NT','Internet Explorer 5.0'),(772,'P3P Client',NULL,NULL),(773,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; CHWIE_NL60)','Windows 98','Internet Explorer 5.5'),(774,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 95; QXW0336o)','Windows 98','Internet Explorer 5.5'),(775,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Win 9x 4.90; FreeSurf myweb.nl v1.0.1)','Windows 98','Internet Explorer 5.5'),(776,'Internet Explorer 5.5',NULL,'Internet Explorer 5.5'),(777,'Mozilla/5.0 (Windows; U; Windows NT 5.0; en-US; rv:0.9.2) Gecko/20010726 Netscape6/6.1','Windows 2000','Netscape 6.1'),(778,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; KITV4 Wanadoo)','Windows 98','Internet Explorer 6.0'),(779,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Win 9x 4.90; Het Net 4.0; H010818)','Windows 98','Internet Explorer 5.5'),(780,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; SKY11; Hotbar 2.0; Q312461)','Windows 98','Internet Explorer 6.0'),(781,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; SURFkit V4)','Windows 98','Internet Explorer 5.5'),(782,'Mozilla/4.0 (compatible; MSIE 5.01; Windows NT 5.0; Tucows)','Windows 2000','Internet Explorer 5.0'),(783,'0',NULL,NULL),(784,'Mozilla/5.0 (Macintosh; N; PPC; en-GB; m18) Gecko/20010131 Netscape6/6.01','Mac OS','Netscape 6.01'),(785,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.0; Q312461; DT)','Windows 2000','Internet Explorer 6.0'),(786,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; FreeSurf myweb.nl v1.0.1)','Windows 98','Internet Explorer 5.5'),(787,'-',NULL,NULL),(788,'Anonymous',NULL,NULL),(789,'Mozilla/4.0 (compatible; MSIE 5.01; Windows 95; DT)','Windows 98','Internet Explorer 5.0'),(790,'lwp-request/2.01',NULL,NULL),(791,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; DigExt; Wanadoo cable)','Windows 98','Internet Explorer 5.0'),(792,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Hotbar 4.0)','Windows 98','Internet Explorer 5.5'),(793,'Mozilla/5.0 (Macintosh; U; PPC; en-US; rv:0.9.4.1) Gecko/20020508 Netscape6/6.2.3','Mac OS','Netscape 6.2'),(794,'Mozilla/4.6 [en] (WinNT; I)','Windows NT','Netscape 4.6'),(795,'Mozilla/5.0 (Slurp/si; slurp@inktomi.com; http://www.inktomi.com/slurp.html)',NULL,NULL),(796,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; c\'t)','Windows XP','Internet Explorer 6.0'),(797,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; DigExt; DVD Owner)','Windows 98','Internet Explorer 5.0'),(798,'Mozilla/4.6 [en-gb]C-CCK-MCD NetscapeOnline.co.uk  (Win98; I)','Windows 98','Netscape 4.6'),(799,'Mozilla/4.0 (compatible; MSIE 5.01; Windows 98; SKY11a)','Windows 98','Internet Explorer 5.0'),(800,'Mozilla/4.5 (compatible; OmniWeb/4.1-v410; Mac_PowerPC)','Mac OS','OmniWeb 4.1'),(801,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; AtHomeNL0107)','Windows 98','Internet Explorer 5.5'),(802,'Mozilla/5.0 (Windows; U; WinNT4.0; en-US; rv:1.0.0) Gecko/20020530','Windows NT 4.0','Mozilla 1.0'),(803,'Mozilla/4.0 (compatible; MSIE 4.0; Windows 98)','Windows 98','Internet Explorer 4.0'),(804,'Mozilla/4.0 (compatible; MSIE 5.01; Windows NT; DigExt)','Windows NT','Internet Explorer 5.0'),(805,'Mozilla/4.0 (compatible; MSIE 5.01; Windows NT; Internet Explorer 5; Windows NT4; Reuters MCD.)','Windows NT','Internet Explorer 5.0'),(806,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; Hotbar 3.0; Planet Internet PIM 2)','Windows 98','Internet Explorer 6.0'),(807,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; wish-00-07)','Windows 98','Internet Explorer 6.0'),(808,'Mozilla/4.08 (Macintosh; I; PPC, Nav)','Mac OS','Netscape 4.08'),(809,'Mozilla/4.0 (compatible; MSIE 5.01; Windows NT 5.0; Hotbar 4.0)','Windows 2000','Internet Explorer 5.0'),(810,'Xenu\'s Link Sleuth 1.1c',NULL,NULL),(811,'Internet Explore 5.x',NULL,NULL),(812,'Mozilla/4.0 (compatible; MSIE 5.01; Windows 95; wish-00-07)','Windows 98','Internet Explorer 5.0'),(813,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 95; DigExt; Planet Internet PIM 2)','Windows 98','Internet Explorer 5.0'),(814,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; Win 9x 4.90; AtHome0107; Q312461)','Windows 98','Internet Explorer 6.0'),(815,'Mozilla/4.0 (compatible; MSIE 4.01; Windows 98; PKBL008)','Windows 98','Internet Explorer 4.01'),(816,'Mozilla/4.0 (compatible; MSIE 5.0; Windows XP) Opera 6.01  [en]','Windows XP','Internet Explorer 5.0'),(817,'Mozilla/4.0 (compatible; MSIE 4.01; Windows 95; Wanadoo NL)','Windows 98','Internet Explorer 4.01'),(818,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; Win 9x 4.90; WanadooNL31)','Windows 98','Internet Explorer 6.0'),(819,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; DigExt; Planet Internet PIM 2)','Windows 98','Internet Explorer 5.0'),(820,'SIE-C3I/3.0 UP/4.1.16m UP.Browser/4.1.16m-XXXX UP.Link/4.2.2.9',NULL,NULL),(821,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; PCAdvisor032002)','Windows XP','Internet Explorer 6.0'),(822,'Mozilla/5.0 (Macintosh; U; PPC; en-US; rv:0.9.2) Gecko/20010728 Netscape6/6.1','Mac OS','Netscape 6.1'),(823,'Mozilla/4.0 (compatible; MSIE 5.2; Mac_PowerPC)','Mac OS','Internet Explorer 5.2'),(824,'libwww-perl/5.63',NULL,NULL),(825,'bumblebee/1.0 (bumblebee@relevare.com; http://www.relevare.com/)',NULL,NULL),(826,'Mozilla/4.5 (compatible; OmniWeb/4.1-v422; Mac_PowerPC)','Mac OS','OmniWeb 4.1'),(827,'Mozilla/4.0 (compatible; MSIE 5.5; Windows NT 4.0; %username%; H010818)','Windows NT 4.0','Internet Explorer 5.5'),(828,'Mozilla/4.0 (compatible; MSIE 5.0; Windows XP) Opera 6.02  [en]','Windows XP','Internet Explorer 5.0'),(829,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; HCCnet Internet Explorer; DigExt)','Windows 98','Internet Explorer 5.0'),(830,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; Planet-ADSL; T312461)','Windows 98','Internet Explorer 6.0'),(831,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.0; MSIECrawler)','Windows 2000','Internet Explorer 6.0'),(832,'Mozilla/4.75 [de] (WinNT; U)','Windows NT','Netscape 4.75'),(833,'Mozilla/4.0 (compatible; MSIE 5.5; Windows NT 4.0; obot)','Windows NT 4.0','Internet Explorer 5.5'),(834,'Mozilla/4.0 (compatible; Powermarks/3.5; Windows 95/98/2000/NT)','Windows 98',NULL),(835,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; {12Move})','Windows 98','Internet Explorer 6.0'),(836,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; Concepts ICT Excellence)','Windows XP','Internet Explorer 6.0'),(837,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Win 9x 4.90; FREELER1.1)','Windows 98','Internet Explorer 5.5'),(838,'Mozilla/5.0 (Macintosh; U; PPC Mac OS X; en-US; rv:1.0.0) Gecko/20020529','Mac OS X','Mozilla 1.0'),(839,'Mozilla/4.5 [nl] (WinNT; I)','Windows NT','Netscape 4.5'),(840,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 95; DigExt; Mozilla/4.0 (Compatibel; MSIE 5.0; Windows NT; Solcon ISP))','Windows 98','Internet Explorer 5.0'),(841,'http://www.almaden.ibm.com/cs/crawler   [c01]',NULL,NULL),(842,'Mozilla/5.0 (Windows; U; Windows NT 5.0; en-US; rv:1.0.0) Gecko/20020530','Windows 2000','Mozilla 1.0'),(843,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; Win 9x 4.90; FreeSurf myweb.nl v1.0.1)','Windows 98','Internet Explorer 6.0'),(844,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; H010818; T312461)','Windows 98','Internet Explorer 5.5'),(845,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.0; BaByXL BroadBand DSL)','Windows 2000','Internet Explorer 6.0'),(846,'Mozilla/4.51 [en]C-ASML_1P06S  (WinNT; I)','Windows NT','Netscape 4.51'),(847,'Mozilla/4.0 (compatible; MSIE 5.0; Windows XP) Opera 6.03  [en]','Windows XP','Internet Explorer 5.0'),(848,'Mozilla/4.5 (compatible; iCab 2.8; Macintosh; I; PPC)','Mac OS',NULL),(849,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; Hotbar 3.0)','Windows 98','Internet Explorer 5.0'),(850,'Vindex 1.3  (Spider 2.1/info@vang.net)',NULL,NULL),(851,'Mozilla/4.0 (compatible; MSIE 5.5; Windows NT 4.0; RuG Bedrijfskunde)','Windows NT 4.0','Internet Explorer 5.5'),(852,'Links (0.95; Unix)','Unix','Links 0.9'),(853,'Links (2.0pre4; Linux 2.4.18-grsec-1.9.4 i686; 83x40)','Linux','Links 2.0'),(854,'Mozilla/4.71 [en] (WinNT; I)','Windows NT','Netscape 4.71'),(855,'Mozilla/2.0 (compatible; MSIE 3.01; Windows 95)','Windows 98','Internet Explorer 3.01'),(856,'Mozilla/5.0 (Windows; U; Win 9x 4.90; en-GB; rv:0.9.2) Gecko/20010726 Netscape6/6.1','Windows ME','Netscape 6.1'),(857,'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:0.9.4.1) Gecko/20020314 Netscape6/6.2.2','Linux','Netscape 6.2'),(858,'Mozilla/3.0 (compatible; StarOffice/5.2;Linux)','Linux',NULL),(859,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; PKBL008; DigExt; KITV4 Wanadoo)','Windows 98','Internet Explorer 5.0'),(860,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; MSOCD; AtHomeNL191)','Windows XP','Internet Explorer 6.0'),(861,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; DigExt; FreeSurf etrade.nl v1.0.1)','Windows 98','Internet Explorer 5.0'),(862,'Mozilla/3.0 (Vagabondo/2.0 MT; webcrawler@NOSPAMwise-guys.nl; http://webagent.wise-guys.nl/)',NULL,NULL),(863,'Gigabot/1.0',NULL,NULL),(864,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; (R1 1.1))','Windows XP','Internet Explorer 6.0'),(865,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; Creative; T312461; Q312461)','Windows XP','Internet Explorer 6.0'),(866,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; DigExt; Planet-ADSL)','Windows 98','Internet Explorer 5.0'),(867,'Mozilla/4.0 (compatible; MSIE 5.01; Windows 98; KA V3 UK)','Windows 98','Internet Explorer 5.0'),(868,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 95; DigExt; Open Universiteit Nederland; installatie CD-rom; FreeSurf tmfweb.nl v1.0.1)','Windows 95','Internet Explorer 5.0'),(869,'Mozilla/4.0 (compatible; MSIE 5.01; Windows 95; {Voetbal NL})','Windows 95','Internet Explorer 5.0'),(870,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; AIRF)','Windows XP','Internet Explorer 6.0'),(871,'Mozilla/5.0 (compatible; MSIE 6.0; Windows XP)','Windows XP','Internet Explorer 6.0'),(872,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; BCD2000)','Windows 98','Internet Explorer 5.5'),(873,'Mozilla/5.0 Galeon/1.2.5 (X11; Linux i686; U;) Gecko/20020606','Linux','Galeon 1.2'),(874,'Zoek.nl Spider',NULL,NULL),(875,'Mozilla/4.0 (compatible; MSIE 5.01; Windows NT; Aztec)',NULL,'Internet Explorer 5.0'),(876,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Win 9x 4.90; Planet Internet PIM 2)','Windows 98','Internet Explorer 5.5'),(877,'psbot/0.1 (+http://www.picsearch.com/bot.html)',NULL,NULL),(878,'Mozilla/4.7 [nl] (WinNT; U)',NULL,'Netscape 4.7'),(879,'Mozilla/4.08 [en] (Win98; I ;Nav)','Windows 98','Netscape 4'),(880,'Mozilla/4.05 [en] (WinNT; I)',NULL,'Netscape 4'),(881,'Java1.3.0',NULL,NULL),(882,'Mozilla/5.0 (Windows; U; Win 9x 4.90; en-US; rv:0.9.2) Gecko/20010726 Netscape6/6.1','Windows ME','Netscape 6.1'),(883,'Scooter_x0-3.2.EX',NULL,NULL),(884,'Scooter-3.2.FNR',NULL,NULL),(885,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; HCCnet)','Windows 98','Internet Explorer 6.0'),(886,'Mozilla/4.0',NULL,NULL),(887,'Mozilla/4.77 [en] (Windows NT 5.0; U)','Windows 2000','Netscape 4.77'),(888,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Win 9x 4.90; .NET CLR 1.0.3705)','Windows 98','Internet Explorer 5.5'),(889,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 95; DigExt; PC SRP 1.5)','Windows 95','Internet Explorer 5.0'),(890,'Mozilla/4.75 [en] (WinNT; U)','Windows NT 4.0','Netscape 4.7'),(891,'Mozilla/4.0 (compatible; MSIE 5.0; Win32)',NULL,'Internet Explorer 5.0'),(892,'Mozilla/4.0 (compatible; MSIE 6.0; Win32)',NULL,'Internet Explorer 6.0'),(893,'Mozilla/4.73 [en] (X11; I; Linux 2.4.0-test3 i686)','Linux','Netscape 4.73'),(894,'Links (2.0pre4; Linux 2.4.18-grsec-1.9.4 i686; 80x25)','Linux','Links 2.0'),(895,'Wget/1.8.1',NULL,'Wget 1.8'),(896,'Microsoft-WebDAV-MiniRedir/5.1.2600',NULL,NULL),(897,'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:0.9.9) Gecko/20020513','Linux','Mozilla 0.9'),(898,'Mozilla/5.0 (X11; U; Linux 2.4.2-2 i686; en-US; 0.7) Gecko/20010316','Linux','Mozilla 0.7'),(899,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 2000) Opera 6.04  [en]','Windows 2000','Opera 6.0'),(900,'Windows-Media-Player/8.00.00.4482',NULL,'Windows Media Player 8'),(901,'fmII URL validator/1.0',NULL,NULL),(902,'Mozilla/5.0 (Windows; U; Windows NT 5.0; en-US; rv:1.1a) Gecko/20020611','Windows 2000','Mozilla 1.1'),(903,'Mozilla/5.0 (Windows; U; Win98; en-US; rv:1.0.0) Gecko/20020530','Windows 98','Mozilla 1.0'),(904,'Mozilla/5.0 (compatible; Konqueror/3; Linux; de, de_CH)','Linux','Konqueror 3'),(905,'Mozilla/5.0 (Windows; U; Win98; en-US; rv:1.0rc3) Gecko/20020523','Windows 98','Mozilla 1.0'),(906,'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.0.0) Gecko/20020615 Debian/1.0.0-3','Linux','Mozilla 1.0'),(907,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 4.0; Q312461; .NET CLR 1.0.3705)','Windows NT 4.0','Internet Explorer 6.0'),(908,'Mozilla/4.0 (compatible; MSIE 5.0; Linux 2.4.2 i386) Opera 6.02  [en]','Linux','Opera 6.0'),(909,'Mozilla/4.75 [en] (X11; U; OpenBSD 3.1 i386)','OpenBSD','Netscape 4.75'),(910,'Lynx/2.8.4dev.8 libwww-FM/2.14 SSL-MM/1.4.1 OpenSSL/0.9.5a',NULL,'Lynx 2.8'),(911,'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.1b) Gecko/20020802','Linux','Mozilla 1.1'),(912,'Mozilla/4.05 (Macintosh; I; PPC)','Mac OS','Netscape 4'),(913,'Mozilla/5.0 (Linux 2.4.2 i386; U) Opera 6.02  [en]','Linux','Opera 6.0'),(914,'Mozilla/5.0 (compatible; Konqueror/3; Linux)','Linux','Konqueror 3'),(915,'Mozilla/5.0 (compatible; Konqueror/3.0.0-10; nl, en_US, en)',NULL,'Konqueror 3.0'),(916,'Mozilla/4.0 (compatible; MSIE 5.0; Linux) Opera 6.0  [en]','Linux','Opera 6.0'),(917,'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.0.0) Gecko/20020605','Linux','Mozilla 1.0'),(918,'Mozilla/4.0 (Linux 2.4.19-rc3)','Linux',NULL),(919,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; Q312461; NetCaptor 7.0.1; .NET CLR 1.0.3705)','Windows XP','Internet Explorer 6.0'),(920,'Mozilla/4.77 [en] (X11; U; Linux 2.4.16 i686)','Linux','Netscape 4.77'),(921,'Mozilla/4.0 (compatible; MSIE 5.00; Windows 98)','Windows 98','Internet Explorer 5.0'),(922,'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.1a) Gecko/20020610','Linux','Mozilla 1.1'),(923,'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.0.0) Gecko/20020529','Linux','Mozilla 1.0'),(924,'Mozilla/4.76 (Linux 2.4.18-5 i686; U) Opera 6.0  [en]','Linux','Opera 6.0'),(925,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; NetCaptor 7.0.1)','Windows XP','Internet Explorer 6.0'),(926,'Mozilla/4.7 [en]C-CCK-MCD   (WinNT; U)','Windows NT 4.0','Netscape 4.7'),(927,'Mozilla/4.79 [en] (Windows NT 5.0; U)','Windows 2000','Netscape 4.79'),(928,'Mozilla/4.77 [en] (X11; U; Linux 2.4.18 i686)','Linux','Netscape 4.77'),(929,'Mozilla/4.78 [ja] (Windows NT 5.0; U)','Windows 2000','Netscape 4.78'),(930,'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.1b) Gecko/20020727','Linux','Mozilla 1.1'),(931,'Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US; rv:1.0.0) Gecko/20020530','Windows XP','Mozilla 1.0'),(932,'Mozilla/5.0 (Windows; U; Windows NT 5.0; en-US; rv:1.0rc1) Gecko/20020417','Windows 2000','Mozilla 1.0'),(933,'Mozilla/5.0 (Windows; U; Windows NT 5.0; en-US; rv:1.1b) Gecko/20020721','Windows 2000','Mozilla 1.1'),(934,'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.0rc3) Gecko/20020602','Linux','Mozilla 1.0'),(935,'Opera/6.02 (Linux 2.4.19 i686; U)  [en]','Linux','Opera 6.0'),(936,'Mozilla/4.0 (compatible; MSIE 5.21; Mac_PowerPC)','Mac OS','Internet Explorer 5.2'),(937,'Mozilla/5.0 Galeon/1.2.5 (X11; Linux i686; U;) Gecko/20020610 Debian/1.2.5-1','Linux','Galeon 1.2'),(938,'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.1b) Gecko/20020722','Linux','Mozilla 1.1'),(939,'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:0.9.9) Gecko/20020516','Linux','Mozilla 0.9'),(940,'Mozilla/4.0 (compatible; MSIE 6.0; Win32; Hotbar 3.0)',NULL,'Internet Explorer 6.0'),(941,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.0; Crazy Browser 1.0.5)','Windows 2000','Internet Explorer 6.0'),(942,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; t20011122h)','Windows 98','Internet Explorer 6.0'),(943,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 2000) Opera 6.03  [en]','Windows 2000','Opera 6.0'),(944,'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.0.0) Gecko/20020510','Linux','Mozilla 1.0'),(945,'Mozilla/5.0 Galeon/1.2.5 (X11; Linux i686; U;) Gecko/0','Linux','Galeon 1.2'),(946,'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.0rc2) Gecko/20020527','Linux','Mozilla 1.0'),(947,'Mozilla/5.0 (compatible; Konqueror/2.1.2; X11)','Linux','Konqueror 2.1'),(948,'Lynx/2.8.3rel.1 libwww-FM/2.14',NULL,'Lynx 2.8'),(949,'Mozilla/5.0 (X11; U; Linux i386; en-US; rv:1.0.0) Gecko/20020529','Linux','Mozilla 1.0'),(950,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; Q312461; Hotbar 4.0)','Windows 98','Internet Explorer 6.0'),(951,'Mozilla/5.0 (Linux 2.4.19-gentoo-r7 i686; U) Opera 6.02  [en]','Linux','Opera 6.0'),(952,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; BTopenworld)','Windows XP','Internet Explorer 6.0'),(953,'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.1b) Gecko/20020724','Linux','Mozilla 1.1'),(954,'Mozilla/5.0 (compatible; Konqueror/3.0.0-10; Linux)','Linux','Konqueror 3'),(955,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; StarBand Version 1.0; .NET CLR 1.0.3705)','Windows 98','Internet Explorer 6.0'),(956,'Dual Proxy',NULL,NULL),(957,'Mozilla/4.77 [en] (X11; U; Linux 2.4.5 i586)','Linux','Netscape 4.77'),(958,'Mozilla/5.0 Galeon/1.2.1 (X11; Linux i686; U;) Gecko/20020421 Debian/1.2.1-2','Linux','Galeon 1.2'),(959,'Java1.4.0',NULL,'Java 1.4'),(960,'Mozilla/5.0 (Macintosh; U; PPC Mac OS X; en-US; rv:1.0rc3) Gecko/20020523','Mac OS X','Mozilla 1.0'),(961,'Mozilla/4.0 (compatible; MSIE 5.01; Windows NT 5.0; H010818; NetCaptor 6.5.0RC1; .NET CLR 1.0.3705)','Windows 2000','Internet Explorer 5.0'),(962,'Mozilla/4.5 (compatible; HTTrack 3.0x; Windows 98)','Windows 98','HTTrack 3.0'),(963,'Mozilla/5.0 (Windows; U; Windows NT 5.0; en-US; rv:1.1a+) Gecko/20020701','Windows 2000','Mozilla 1.1'),(964,'fetch libfetch/2.0',NULL,NULL),(965,'Mozilla/4.0 (compatible; MSIE 5.01; Windows NT 6.0; Version; Compatible)',NULL,'Internet Explorer 5.0'),(966,'Opera/6.04 (Windows XP; U)  [en]','Windows XP','Opera 6.0'),(967,'Links (2.0pre4; Linux 2.4.18-grsec-1.9.4 i686; 80x30)','Linux','Links 2.0'),(968,'Mozilla/5.0 (compatible; Konqueror/3.0.0; FreeBSD)','FreeBSD','Konqueror 3'),(969,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; VV6q0jjDXnZFBv6.)','Windows 98','Internet Explorer 6.0'),(970,'GetRight/4.5d',NULL,'GetRight 4.5'),(971,'Mozilla/4.78 [en] (X11; U; Linux 2.4.7-10smp i686)','Linux','Netscape 4.78'),(972,'Mozilla/5.0 (Windows; U; Windows NT 5.0; de-DE; rv:0.9.4.1) Gecko/20020314 Netscape6/6.2.2','Windows 2000','Netscape 6.2'),(973,'Mozilla/5.0 (Windows 98; U) Opera 6.04  [en]','Windows 98','Opera 6.0'),(974,'Mozilla/4.05 [en] (Win98; I))','Windows 98','Netscape 4'),(975,'Mozilla/5.0 (X11; U; Linux i686; zh-TW; rv:1.0.0) Gecko/20020615 Debian/1.0.0-3','Linux','Mozilla 1.0'),(976,'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.0.0) Gecko/20020724','Linux','Mozilla 1.0'),(977,'Mozilla/5.0 (Windows; U; Windows NT 5.0; fr-FR; rv:1.0.0) Gecko/20020530','Windows 2000','Mozilla 1.0'),(978,'Mozilla/4.0 (compatible; MSIE 5.5; Windows NT 4.0; NetCaptor 7.0 Beta 5; Crazy Browser 1.0.5)','Windows NT 4.0','Internet Explorer 5.5'),(979,'NG/1.0',NULL,NULL),(980,'Java1.3.1',NULL,'Java 1.3'),(981,'Mozilla/5.0 (Windows; U; Windows NT 5.0; en-US; rv:1.1b) Gecko/20020801','Windows 2000','Mozilla 1.1'),(982,'Industry Update Control',NULL,NULL),(983,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; Q312461; Hotbar 3.0; .NET CLR 1.0.3705; MSIECrawler)','Windows XP','Internet Explorer 6.0'),(984,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; MSIECrawler)','Windows XP','Internet Explorer 6.0'),(985,'Mozilla/5.0 (Windows; U; Windows NT 5.0; nl-NL; rv:1.0.0) Gecko/20020530','Windows 2000','Mozilla 1.0'),(986,'Mozilla/5.0 (Linux 2.4.18-bfh i686; U) Opera 5.0  [en]','Linux','Opera 5.0'),(987,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.0; Q312461; .NET CLR 1.0.3705)','Windows 2000','Internet Explorer 6.0'),(988,'Mozilla/5.0 Galeon/1.0.2 (X11; Linux i586; U;) Gecko/20011226','Linux','Galeon 1.0'),(989,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; MSIECrawler)','Windows 98','Internet Explorer 6.0'),(990,'FreeFind.com-SiteSearchEngine/1.0 (http://freefind.com; provides search engines to web sites; spiderinfo@freefind.com)',NULL,NULL),(991,'Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US; rv:0.9.4.1) Gecko/20020508 Netscape6/6.2.3','Windows XP','Netscape 6.2'),(992,'Mozilla/4.0 (compatible; MSIE 5.0; Windows XP) Opera 6.04  [en]','Windows XP','Opera 6.0'),(993,'Links (0.1; Linux 2.4.12 i686)','Linux','Links 0.1'),(994,'URLBlaze',NULL,NULL),(995,'Mozilla/5.0 (compatible; Konqueror/2.2.2; Linux)','Linux','Konqueror 2.2.2'),(996,'Mozilla/4.78 [en] (X11; U; Linux 2.4.18-6mdk i686)','Linux','Netscape 4.78'),(997,'Mozilla/4.77 [en] (X11; U; Linux 2.4.9 i686)','Linux','Netscape 4.77'),(998,'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:0.9.4.1) Gecko/20020508 Netscape6/6.2.3','Linux','Mozilla 0.9'),(999,'Mozilla/4.0 (compatible; MSIE 5.0; Windows XP) Opera 6.01  [fr]','Windows XP','Opera 6.0'),(1000,'Microsoft BITS/6.2',NULL,NULL),(1001,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; *** TANABATA ***; T312461)','Windows 98','Internet Explorer 5.5'),(1002,'Mozilla/5.0 Galeon/1.2.5 (X11; Linux i686; U;) Gecko/20020605','Linux','Galeon 1.2'),(1003,'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.0.0) Gecko/00200205','Linux','Mozilla 1.0'),(1004,'Mozilla/4.0 (compatible; MSIE 5.0; Linux) Opera 6.02  [en]','Linux','Opera 6.0'),(1005,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; Hotbar 4.0)','Windows 98','Internet Explorer 6.0'),(1006,'Wget/1.8.2',NULL,'Wget 1.8'),(1007,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.0; H010818; .NET CLR 1.0.3705)','Windows 2000','Internet Explorer 6.0'),(1008,'Mozilla/4.78 (Windows 2000; U) Opera 6.04  [en]','Windows 2000','Opera 6.0'),(1009,'Mozilla/4.0 (compatible; MSIE 5.15; Mac_PowerPC)','Mac OS','Internet Explorer 5.1'),(1010,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; T312461; Hotbar 3.0; .NET CLR 1.0.3705)','Windows 98','Internet Explorer 6.0'),(1011,'Mozilla/5.0 Galeon/1.0.3 (X11; Linux i686; U;) Gecko/20020325','Linux','Galeon 1.0'),(1012,'Windows-Media-Player/8.00.00.4477',NULL,'Windows Media Player 8'),(1013,'Mozilla/5.0 (X11; U; FreeBSD i386; en-US; rv:1.0rc2) Gecko/20020812','FreeBSD','Mozilla 1.0'),(1014,'DeepIndexer',NULL,NULL),(1015,'Opera/6.0 (Linux 2.4.18-4GB i586; U)  [en]','Linux','Opera 6.0'),(1016,'Mozilla/4.0 (compatible; MSIE 5.0; Windows XP) Opera 6.03  [ja]','Windows XP','Opera 6.0'),(1017,'Opera/6.04 (Windows XP; U) ~ [~ [nl]','Windows XP','Opera 6.0'),(1018,'Mozilla/4.0 compatible ZyBorg/1.0 (wn.zyborg@looksmart.net; http://www.WISEnutbot.com)',NULL,NULL),(1019,'DA 5.3',NULL,NULL),(1020,'Mozilla/4.0 (Windows XP 5.1)','Windows XP',NULL),(1021,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; DVD Owner)','Windows XP','Internet Explorer 6.0'),(1022,'Mozilla/5.0 (Macintosh; U; PPC Mac OS X; en-US; rv:1.1b) Gecko/20020722','Mac OS X','Mozilla 1.1'),(1023,'ELinks (0.4pre12.CVS; Linux 2.4.18-21mdkcustom i686; 128x48)','Linux','ELinks 0.4'),(1024,'Mozilla/4.0 (compatible; MSIE 5.0; Linux 2.4.18-3 i686) Opera 6.02  [en]','Linux','Opera 6.0'),(1025,'W3C_Validator/1.183 libwww-perl/5.64',NULL,NULL),(1026,'Mozilla/5.0 (X11; U; FreeBSD i386; en-US; rv:1.0rc3) Gecko/20020607','FreeBSD','Mozilla 1.0'),(1027,'Mozilla/5.0 (compatible; Konqueror/3.0.0; Linux)','Linux','Konqueror 3'),(1028,'Links (0.96; Unix)','Unix','Links 0.9'),(1029,'Mozilla/4.79 [de] (X11; U; Linux 2.4.18 i686)','Linux','Netscape 4.79'),(1030,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; DigExt; fs_pb_ie5)','Windows 98','Internet Explorer 5.0'),(1031,'Mozilla/4.0 (compatible; MSIE 5.01; Windows NT 5.0) / StripIt 0.2','Windows 2000','Internet Explorer 5.0'),(1032,'Mozilla/5.0 (Windows; U; Win98; en-US; rv:0.9.4) Gecko/20011019 Netscape6/6.2 / StripIt 0.2','Windows 98','Netscape 6.2'),(1033,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; DigExt) / StripIt 0.2','Windows 98','Internet Explorer 5.0'),(1034,'Mozilla/5.0 (Windows; U; Windows NT 5.0; en-US; rv:1.0.1) Gecko/20020823','Windows 2000','Mozilla 1.0'),(1035,'Mozilla/5.0 (Macintosh; U; PPC Mac OS X; en-US; rv:1.1) Gecko/20020826','Mac OS X','Mozilla 1.1'),(1036,'Mozilla/4.0 (compatible; MSIE 5.5; AOL 7.0; Windows 98; DT)','Windows 98','Internet Explorer 5.5'),(1037,'Mozilla/5.0 Galeon/1.2.0 (X11; Linux i686; U;) Gecko/20020408','Linux','Galeon 1.2'),(1038,'RMA/1.0 (compatible; RealMedia)',NULL,NULL),(1039,'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.0.1) Gecko/20020823 Netscape/7.0','Linux','Netscape 7.0'),(1040,'Mozilla/4.0 (compatible; MSIE 5.0; Windows XP) Opera 6.05  [en]','Windows XP','Opera 6.0'),(1041,'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.1) Gecko/20020826','Linux','Mozilla 1.1'),(1042,'Mozilla/4.0 (compatible; MSIE 5.5; Windows NT 5.0; H010818; MyIE2 0.3)','Windows 2000','Internet Explorer 5.5'),(1043,'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.0rc1) Gecko/20020417','Linux','Mozilla 1.0'),(1044,'Java1.3.1_02',NULL,'Java 1.3'),(1045,'Mozilla/4.77 [en] (Win98; U)','Windows 98','Netscape 4.77'),(1046,'Mozilla/5.0 (Windows; U; Win98; en-US; rv:1.1) Gecko/20020826','Windows 98','Mozilla 1.1'),(1047,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; MAK Productions; Hogeschool Haarlem PABO)','Windows 98','Internet Explorer 6.0'),(1048,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.0; DT)','Windows 2000','Internet Explorer 6.0'),(1049,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.0; 1.21.2 )','Windows 2000','Internet Explorer 6.0'),(1050,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; DigExt; NL-MultiWebUser)','Windows 98','Internet Explorer 5.0'),(1051,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 2000) Opera 6.05  [en]','Windows 2000','Opera 6.0'),(1052,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; DigExt; YComp 5.0.0.0)','Windows 98','Internet Explorer 5.0'),(1053,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Win 9x 4.90; YComp 5.0.0.0)','Windows 98','Internet Explorer 5.5'),(1054,'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.0.0) Gecko/20020606','Linux','Mozilla 1.0'),(1055,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Win 9x 4.90; MAK Productions)','Windows 98','Internet Explorer 5.5'),(1056,'Mozilla/5.0 (X11; U; Linux 2.4.18-grsec-1.9.4 i686; en-US; rv:0.9.1) Gecko/20010610','Linux','Mozilla 0.9'),(1057,'Mozilla/5.0 (Windows; U; Windows NT 5.0; en-US; rv:1.2a) Gecko/20020910','Windows 2000','Mozilla 1.2'),(1058,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.0; YComp 5.0.0.0)','Windows 2000','Internet Explorer 6.0'),(1059,'Netscape Navigator 4.05',NULL,'Netscape 4.0'),(1060,'Mozilla/5.0 Galeon/1.2.5 (X11; Linux ppc; U;) Gecko/20020610 Debian/1.2.5-1','Linux','Galeon 1.2'),(1061,'Mozilla/4.0 (compatible; MSIE 5.5; Windows NT 5.0; T312461; Lunascape 0.99beta5)','Windows 2000','Internet Explorer 5.5'),(1062,'Mozilla/5.0 (compatible; MSIE 5.0; <-- ha ha ha, yeah...right)',NULL,''),(1063,'Mozilla/4.0 (compatible; MSIE 5.01; Windows NT; YComp 5.0.0.0)','Windows NT 4.0','Internet Explorer 5.0'),(1064,'Mozilla/4.78 (TuringOS; Turing Machine; 0.0)',NULL,NULL),(1065,'Lynx/2.8.2rel.1 libwww-FM/2.14',NULL,'Lynx 2.8'),(1066,'Mozilla/4.0 (compatible; MSIE 5.01; Windows 98; .NET CLR 1.0.3705)','Windows 98','Internet Explorer 5.0'),(1067,'Opera/6.03 (Linux 2.4.19 i686; U)  [en]','Linux','Opera 6.0'),(1068,'Wget/1.8',NULL,'Wget 1.8'),(1069,'Mozilla/4.0 (compatible; MSIE 5.01; Windows NT 5.0; DigExt; Hotbar 3.0)','Windows 2000','Internet Explorer 5.0'),(1070,'libwww-perl/5.48',NULL,NULL),(1071,'Mozilla/4.0 (compatible; MSIE 5.0; Linux 2.4.19-bfh i686) Opera 5.0  [en]','Linux','Opera 5.0'),(1072,'Mozilla/4.0 (compatible; MSIE 5.0; Linux) Opera 5.0  [en]','Linux','Opera 6.0'),(1073,'Mozilla/5.0 Galeon/1.0.3 (X11; Linux i686; U;) Gecko/0','Linux','Galeon 1.0'),(1074,'Mozilla/4.76 [en] (Windows NT 5.0; U)','Windows 2000','Netscape 4.76'),(1075,'Links (2.0pre6; Linux 2.4.18-10asp i686; 99x31)','Linux','Links 2.0'),(1076,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; MSOCD; AtHomeNL191; YComp 5.0.0.0; Hotbar 3.0)','Windows 98','Internet Explorer 5.5'),(1077,'Outlook-Express/6.0 (MSIE 6.0; Windows NT 5.0; YComp 5.0.0.0; TmstmpExt)','Windows 2000','Internet Explorer 6.0'),(1078,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; DigExt; Planet Internet PIM 2; YComp 5.0.2.4)','Windows 98','Internet Explorer 5.0'),(1079,'Vagabondo/2.0 MT (webagent at wise-guys dot nl)',NULL,NULL),(1080,'Mozilla/5.0 (compatible; Vagabondo/2.1; webcrawler at wise-guys dot nl; http://webagent.wise-guys.nl/)',NULL,NULL),(1081,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; Win 9x 4.90; SC/5.55/1.70/HP-FriSurf)','Windows 98','Internet Explorer 6.0'),(1082,'Mozilla/4.0 (compatible; MSIE 5.01; Windows 98; KITV4 Wanadoo)','Windows 98','Internet Explorer 5.0'),(1083,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98) via <B>aVirt Gateway Server</B> v3.5','Windows 98','Internet Explorer 6.0'),(1084,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; YComp 5.0.0.0)','Windows XP','Internet Explorer 6.0'),(1085,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; (R1 1.3))','Windows XP','Internet Explorer 6.0'),(1086,'Mozilla/5.0 (Windows; U; Windows NT 5.1; zh-TW; rv:1.1) Gecko/20020826','Windows XP','Mozilla 1.1'),(1087,'Windows-Media-Player/9.00.00.2799',NULL,'Windows Media Player 9'),(1088,'Windows-Media-Player/7.01.00.3055',NULL,'Windows Media Player 7'),(1089,'Download',NULL,NULL),(1090,'Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US; rv:1.0.1) Gecko/20020823 Netscape/7.0','Windows XP','Netscape 7.0'),(1091,'Mozilla/5.0 Galeon/1.2.5 (X11; Linux i686; U;) Gecko/20020809','Linux','Galeon 1.2'),(1092,'Windows-Media-Player/9.00.00.2800',NULL,'Windows Media Player 9'),(1093,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; YComp 5.0.0.0)','Windows 98','Internet Explorer 6.0'),(1094,'Links (2.0pre4; Linux 2.4.19-grsec i686; 80x30)','Linux','Links 2.0'),(1095,'Mozilla/5.0 (Windows; U; Windows NT 5.0; en-US; rv:1.0.1) Gecko/20020826','Windows 2000','Mozilla 1.0'),(1096,'Lynx/2.8.5dev.3 libwww-FM/2.14 SSL-MM/1.4.1 OpenSSL/0.9.6c',NULL,'Lynx 2.8'),(1097,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 95; DigExt; Het Net 3.2)','Windows 95','Internet Explorer 5.0'),(1098,'Mozilla/4.0 (compatible; MSIE 5.22; Mac_PowerPC)','Mac OS','Internet Explorer 5.2'),(1099,'Mozilla/2.0 (compatible; MSIE 3.0; Windows 3.1)','Windows 3.1','Internet Explorer 3.0'),(1100,'Mozilla/4.06 (Win95; I)','Windows 95','Netscape 4'),(1101,'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.1) Gecko/20020927','Linux','Mozilla 1.1'),(1102,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; AtHomeNL191; DVD Owner)','Windows 98','Internet Explorer 5.5'),(1103,'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.0.1) Gecko/20020830','Linux','Mozilla 1.0'),(1104,'Mozilla/4.79 [en] (X11; U; Linux 2.4.18-3 i686)','Linux','Netscape 4.79'),(1105,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; Q312461; Hotbar 4.0)','Windows XP','Internet Explorer 6.0'),(1106,'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.2a) Gecko/20020910','Linux','Mozilla 1.2'),(1107,'Mozilla/4.51 [sv] (Win98; U)','Windows 98','Netscape 4.5'),(1108,'Links (2.0pre4; Linux 2.4.19-grsec i686; 83x42)','Linux','Links 2.0'),(1109,'Mozilla/5.0 Galeon/1.2.5 (X11; Linux i686; U;) Gecko/20020623 Debian/1.2.5-0.woody.1','Linux','Galeon 1.2'),(1110,'Mozilla/5.0 (Windows; U; Windows NT 5.0; en-US; rv:1.1) Gecko/20020826','Windows 2000','Mozilla 1.1'),(1111,'Mozilla/4.0 (compatible; MSIE 5.01; Windows NT 5.0; wguquqS.taLgzOCy; aGrqLdNepaDXKYEQ)','Windows 2000','Internet Explorer 5.0'),(1112,'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.2b) Gecko/20021008 Phoenix/0.3','Linux','Mozilla 1.2'),(1113,'Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US; rv:1.2a) Gecko/20020910','Windows XP','Mozilla 1.2'),(1114,'Star Downloader',NULL,NULL),(1115,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 95; sunrise free surf; DigExt)','Windows 95','Internet Explorer 5.0'),(1116,'NSPlayer/8.0.0.4477',NULL,NULL),(1117,'Mozilla/4.78 [en] (X11; U; OpenBSD 3.0 i386)','OpenBSD','Netscape 4.78'),(1118,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; Hotbar 4.1.2.0)','Windows 98','Internet Explorer 6.0'),(1119,'Microsoft Internet Explorer; Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1)','Windows XP','Internet Explorer 6.0'),(1120,'Windows-Media-Player/8.00.00.4487',NULL,'Windows Media Player 8'),(1121,'TurnitinBot/1.5 (http://www.turnitin.com/robot/crawlerinfo.html)',NULL,NULL),(1122,'TurnitinBot/1.5 http://www.turnitin.com/robot/crawlerinfo.html',NULL,NULL),(1123,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; Hotbar 4.1.8.0)','Windows XP','Internet Explorer 6.0'),(1124,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; Wanadoo01RT43NL)','Windows XP','Internet Explorer 6.0'),(1125,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; NetCaptor 6.5.0B7)','Windows 98','Internet Explorer 6.0'),(1126,'ELinks (0.4pre17; Linux 2.4.10-4GB i686; 111x37)','Linux','ELinks 0.4'),(1127,'Mozilla/5.0 (Windows; U; Windows NT 5.0; en-US; rv:1.2b) Gecko/20021016','Windows 2000','Mozilla 1.2'),(1128,'Lynx/2.8.4rel.1 libwww-FM/2.14',NULL,'Lynx 2.8'),(1129,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; MAK Productions; Hogeschool Haarlem PABO; YComp 5.0.0.0)','Windows 98','Internet Explorer 6.0'),(1130,'Opera/6.03 (Linux 2.4.19 i586; U)  [en]','Linux','Opera 6.0'),(1131,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; DigExt; {Tiscali})','Windows XP','Internet Explorer 6.0'),(1132,'SurveyBot/2.2 <a href=\'http://www.whois.sc\'>Whois Source</a>',NULL,NULL),(1133,'Mozilla/4.0 (compatible; MSIE 5.5; Windows NT 5.0; DT)','Windows 2000','Internet Explorer 5.5'),(1134,'libwww-perl/5.50',NULL,NULL),(1135,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; Multikabel)','Windows XP','Internet Explorer 6.0'),(1136,'Teleport Pro/1.29.1847',NULL,NULL),(1137,'Links (0.96; Linux 2.4.18-17.7asp i686)','Linux','Links 0.9'),(1138,'Mozilla/5.0 (X11; U; Linux i686; ru-RU; rv:1.0.0) Gecko/20020610','Linux','Mozilla 1.0'),(1139,'Elinks (0.3; Linux 2.4.18-3 i586)','Linux','Elinks 0.3'),(1140,'Mozilla/4.0 (compatible; MSIE 5.0; Linux 2.4.10-4GB i686) Opera 5.0  [en]','Linux','Opera 5.0'),(1141,'Lynx/2.8.5dev.7 libwww-FM/2.14 SSL-MM/1.4.1 OpenSSL/0.9.6b',NULL,'Lynx 2.8'),(1142,'/ libwww/5.3.2',NULL,NULL),(1143,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Win 9x 4.90; Hotbar 4.1.8.0)','Windows 98','Internet Explorer 5.5'),(1144,'Mozilla/5.0 (Windows; U; Windows NT 5.0; en-US; rv:1.0.1) Gecko/20020823 Netscape/7.0','Windows 2000','Mozilla 1.0'),(1145,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; DigExt; AtHome033)','Windows 98','Internet Explorer 5.0'),(1146,'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.0rc2) Gecko/20020508','Linux','Mozilla 1.0'),(1147,'Mozilla/5.0 Galeon/1.2.6 (X11; Linux i686; U;) Gecko/20020913 Debian/1.2.6-2','Linux','Galeon 1.2'),(1148,'Scooter/3.2',NULL,'Scooter 3.2'),(1149,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; DigExt; Hi Way, powered by Planet Internet; ZON)','Windows 98','Internet Explorer 5.0'),(1150,'Links (2.0pre4; Linux 2.4.19-grsec i686; 65x34)','Linux','Links 2.0'),(1151,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; DigExt; FreeSurf myweb.nl v1.0.1; YComp 5.0.2.4)','Windows 98','Internet Explorer 5.0'),(1152,'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.1) Gecko/20020918 Debian/1.1-1','Linux','Mozilla 1.1'),(1153,'Baiduspider+(+http://www.baidu.com/search/spider.htm)',NULL,NULL),(1154,'WebCopier v2.7',NULL,'WebCopier 2.7'),(1155,'Mozilla/5.0 (compatible; Konqueror/3)',NULL,'Konqueror 3.0'),(1156,'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.2b) Gecko/20021025','Linux','Mozilla 1.2'),(1157,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.0; T312461; (R1 1.1); .NET CLR 1.0.3705)','Windows 2000','Internet Explorer 6.0'),(1158,'Mozilla/5.0 (Windows; U; Windows NT 5.1; en-CA; rv:1.0.1) Gecko/20020823 Netscape/7.0','Windows XP','Mozilla 1.0'),(1159,'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.2b) Gecko/20021029 Phoenix/0.4','Linux','Mozilla 1.2'),(1160,'Mozilla/4.0 (compatible; MSIE 5.0; Linux 2.4.19 i686) Opera 6.03  [en]','Linux','Opera 6.0'),(1161,'SpaceBison/0.01 [fu] (Win67; X; ShonenKnife)',NULL,NULL),(1162,'Mozilla/5.0 Galeon/1.2.6 (X11; Linux i686; U;) Gecko/20021108','Linux','Galeon 1.2'),(1163,'Mozilla/5.0 Galeon/1.2.6 (X11; Linux i686; U;) Gecko/20020830','Linux','Galeon 1.2'),(1164,'Mozilla/5.0 (Windows; U; Windows NT 5.0; en-US; rv:1.2b) Gecko/20021029 Phoenix/0.4','Windows 2000','Mozilla 1.2'),(1165,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.0; Neostrada Plus 5.4)','Windows 2000','Internet Explorer 6.0'),(1166,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; YComp 5.0.0.0; .NET CLR 1.0.3705)','Windows XP','Internet Explorer 6.0'),(1167,'Mozilla/5.0 (Windows; U; Windows NT 5.0; de-DE; rv:1.0.1) Gecko/20020823 Netscape/7.0','Windows 2000','Netscape 7.0'),(1168,'Opera/6.1 (Linux 2.4.18-17.7asp i686; U)  [en]','Linux','Opera 6.1'),(1169,'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.0.1) Gecko/20020903','Linux','Mozilla 1.0'),(1170,'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.1) Gecko/20020828','Linux','Mozilla 1.1'),(1171,'Windows-Media-Player/7.10.00.3068',NULL,'Windows Media Player 7'),(1172,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; DigExt; T-Online Internatinal AG; .NET CLR 1.0.3705)','Windows XP','Internet Explorer 6.0'),(1173,'Mozilla/5.0 Galeon/1.2.6 (X11; Linux i686; U;) Gecko/20020829','Linux','Galeon 1.2'),(1174,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; CWIE55NL60; CHWIE_NL60)','Windows XP','Internet Explorer 6.0'),(1175,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; T312461; .NET CLR 1.0.3705)','Windows XP','Internet Explorer 6.0'),(1176,'Java 1.1',NULL,'Java 1.1'),(1177,'NSPlayer/7.1.0.3055',NULL,NULL),(1178,'Mozilla/3.01Gold (Macintosh; I; 68K)','Mac OS',NULL),(1179,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; {VIZZAVI0101})','Windows XP','Internet Explorer 6.0'),(1180,'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.0.1) Gecko/20021003','Linux','Mozilla 1.0'),(1181,'DA 5.0',NULL,NULL),(1182,'Mozilla/4.0 (compatible; MSIE 6.0; Win32; Hotbar 4.1.8.0)',NULL,'Internet Explorer 6.0'),(1183,'Mozilla/4.03 [en] (WinNT; I ;Nav)','Windows NT 4.0','Netscape 4.0'),(1184,'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.2b) Gecko/20021017','Linux','Mozilla 1.2'),(1185,'Mozilla/4.0 (compatible; MSIE 5.5; Windows NT 5.0; H010818; T312461; WMM)','Windows 2000','Internet Explorer 5.5'),(1186,'Mozilla/4.0 (compatible; MSIE 5.0; Linux 2.4.18-686 i686) Opera 6.0  [en]','Linux','Opera 6.0'),(1187,'EasyDL/3.02',NULL,NULL),(1188,'Mozilla/5.0 Galeon/1.2.6 (X11; Linux i686; U;) Gecko/0','Linux','Galeon 1.2'),(1189,'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.1) Gecko/20021102','Linux','Mozilla 1.1'),(1190,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 95; Tucows)','Windows 95','Internet Explorer 5.5'),(1191,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; MyIE2)','Windows 98','Internet Explorer 5.5'),(1192,'ASPSeek/1.2.8',NULL,NULL),(1193,'MMHttp',NULL,NULL),(1194,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; FREELER1.1; T312461)','Windows 98','Internet Explorer 5.5'),(1195,'Lynx/2.8.4rel.1 libwww-FM/2.14 SSL-MM/1.4.1 OpenSSL/0.9.4',NULL,'Lynx 2.8'),(1196,'Mozilla/5.0 (compatible; Konqueror/3; Linux; X11; i686; en_GB, de, de_DE@euro)','Linux','Konqueror 3.0'),(1197,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; ZON)','Windows XP','Internet Explorer 6.0'),(1198,'Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US; rv:1.0.1) Gecko/20020826','Windows XP','Mozilla 1.0'),(1199,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; PKBL008; MSOCD; AtHomeNL191)','Windows 98','Internet Explorer 5.5'),(1200,'FAST-WebCrawler/2.2-pre41 MP3',NULL,NULL),(1201,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; Zon Breedband)','Windows XP','Internet Explorer 6.0'),(1202,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Terminal; T312461)','Windows 98','Internet Explorer 5.5'),(1203,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; DigExt; KITV4 Wanadoo)','Windows 98','Internet Explorer 5.0'),(1204,'Mozilla/5.0 (Macintosh; U; PPC Mac OS X; en-US; rv:0.9.4.1) Gecko/20020508 Netscape6/6.2.3 OmniWeb/4.1.1-v424.6','Mac OS','Netscape 6.2'),(1205,'Mozilla/5.0 (Macintosh; U; PPC Mac OS X; en-US; rv:1.0.1) Gecko/20021104 Chimera/0.6','Mac OS','Mozilla 1.0'),(1206,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Win 9x 4.90; YComp 5.0.2.5)','Windows 98','Internet Explorer 5.5'),(1207,'Mozilla/4.0 (Windows XP 5.1) Java/1.4.1-rc','Windows XP','Java 1.4'),(1208,'Lycos-Music-Xml-Fetcher',NULL,NULL),(1209,'Mozilla/4.0 (compatible; MSIE 5.01; Windows NT 5.0; Freegates:youri)','Windows 2000','Internet Explorer 5.0'),(1210,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; YComp 5.0.2.4)','Windows XP','Internet Explorer 6.0'),(1211,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; YComp 5.0.2.5)','Windows XP','Internet Explorer 6.0'),(1212,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; WWP; .NET CLR 1.0.3705)','Windows XP','Internet Explorer 6.0'),(1213,'Mozilla/5.0 Galeon/1.2.6 (X11; Linux i686; U;) Gecko/20020913 Debian/1.2.6-1','Linux','Galeon 1.2'),(1214,'Mozilla/5.0 (compatible; Konqueror/3.1; Linux 2.4.19-alt0.8-up; X11; i686; ru_RU.CP1251, ru_RU, ru)','Linux','Konqueror 3.1'),(1215,'./httpdtype',NULL,NULL),(1216,'Mozilla/5.0 (Macintosh; U; PPC Mac OS X; en-US; rv:1.2.1) Gecko/20021130','Mac OS','Mozilla 1.2'),(1217,'Mozilla/4.0 (compatible; MSIE 5.5; Windows NT 5.0; MyIE2)','Windows 2000','Internet Explorer 5.5'),(1218,'Mozilla/5.0 (Windows; U; Windows NT 5.0; en-US; rv:1.3a) Gecko/20021126','Windows 2000','Mozilla 1.3'),(1219,'Java1.3.1_01',NULL,'Java 1.3'),(1220,'Mozilla/4.0 (Windows XP 5.1) Java/1.4.1_01','Windows XP','Java 1.4'),(1221,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; DigExt; ZON)','Windows XP','Internet Explorer 6.0'),(1222,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; YComp 5.0.2.6)','Windows XP','Internet Explorer  6.0'),(1223,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; YComp 5.0.2.6)','Windows 98','Internet Explorer 6.0'),(1224,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; DigExt; BTopenworld)','Windows 98','Internet Explorer 5.5'),(1225,'Mozilla/2.0 (compatible; Ask Jeeves/Teoma)',NULL,'Ask Jeeves'),(1226,'Mozilla/4.0 (compatible; MSIE 5.5; Windows NT 5.0; T312461; F-5.5SP2-20011022; F-5.0.1-20000221)','Windows 2000','Internet Explorer 5.5'),(1227,'Mozilla/5.0 (X11; U; Linux i686; de-AT; rv:1.0.0) Gecko/20020529','Linux','Mozilla 1.0'),(1228,'Mozilla/4.7 ( compatible; [dk]; Windows NT4.0; DigiExt )','Windows NT 4.0','Netscape 4.7'),(1229,'Mozilla/4.0 ( compatible; MSIE 5.5; Windows 95; NetCaptor )','Windows 95','Internet Explorer 5.5'),(1230,'Mozilla/4.0 ( compatible; [de]; Windows 98; DigiExt )','Windows 98','Netscape 4.0'),(1231,'Mozilla/3.01 ( compatible; MSIE 5.5; AOL 5.0; DigiExt )',NULL,'Internet Explorer 5.5'),(1232,'Mozilla/4.72 ( compatible; MSIE 5.0; Windows NT4.0; FREEI v2.53 )','Windows NT 4.0','Internet Explorer 5.0'),(1233,'Mozilla/3.01 ( compatible; MSIE 4.0; Windows NT5.0; TWRAITH )','Windows 2000',NULL),(1234,'Mozilla/4.72 ( compatible; [jp]; Windows 98; DigiExt )','Windows 98','Netscape 4.7'),(1235,'Mozilla/3.01 ( compatible; MSIE 5.0; AOL 5.0; DigiExt )',NULL,'Internet Explorer 5.0'),(1236,'Mozilla/4.6 ( compatible; MSIE 5.01; Windows NT4.0; DigiExt )','Windows NT 4.0','Internet Explorer 5.0'),(1237,'Mozilla/4.72 ( compatible; MSIE 5.0; Windows NT5.0; MSNIA )','Windows 2000','Internet Explorer 5.0'),(1238,'Mozilla/4.6 ( compatible; [jp]; Windows 95; DigiExt )','Windows 95','Netscape 4.6'),(1239,'Mozilla/4.6 ( compatible; MSIE 5.5; AOL 5.0; ezn IE )',NULL,'Internet'),(1240,'Mozilla/4.73 ( compatible; MSIE 5.5; Windows NT5.0; athome020 )','Windows 2000','Internet Explorer 5.5'),(1241,'Mozilla/3.01 ( compatible; MSIE 5.5; Windows 98; Compaq )','Windows 98','Internet Explorer 5.5'),(1242,'Mozilla/4.72 ( compatible; [jp]; Windows NT5.0; DigiExt )','Windows 2000','Netscape 4.7'),(1243,'Mozilla/4.6 ( compatible; [dk]; Windows 98; DigiExt )','Windows 98','Netscape 4.6'),(1244,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.0; Wanadoo cable)','Windows 2000','Internet Explorer 6.0'),(1245,'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.2) Gecko/20021202','Linux','Mozilla 1.2'),(1246,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; Win 9x 4.90; MSOCD; AtHomeNL191)','Windows 98','Internet Explorer 6.0'),(1247,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; MSN 6.1; MSNbMSFT; MSNmde-at; MSNcOTH)','Windows XP','Internet Explorer 6.0'),(1248,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; HCCnet)','Windows XP','Internet Explorer 6.0'),(1249,'Mozilla/4.0 (compatible; MSIE 5.01; Windows NT 5.0; BWCH-OL-DE-6)','Windows 2000','Internet Explorer 5.0'),(1250,'XupiterToolbar',NULL,NULL),(1251,'Mozilla/5.0 (Windows 98; U) Opera 6.05  [en]','Windows 98','Opera 6.0'),(1252,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.0; TUCOWS)','Windows 2000','Internet Explorer 6.0'),(1253,'Mozilla/5.0 (Windows; U; Windows NT 5.0; de-AT; rv:1.2) Gecko/20021126','Windows 2000','Mozilla 1.2'),(1254,'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.3a) Gecko/20021112','Linux','Mozilla 1.3'),(1255,'Lycos-News-Xml-Fetcher',NULL,NULL),(1256,'Microsoft Data Access Internet Publishing Provider Cache Manager',NULL,NULL),(1257,'Microsoft Data Access Internet Publishing Provider DAV 1.1',NULL,NULL),(1258,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; www.STRATO.de Edition 1.0)','Windows 98','Internet Explorer 6.0'),(1259,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; MSN 6.1; MSNbMSFT; MSNmnl-nl; MSNc00; v5m)','Windows 98','Internet Explorer 6.1'),(1260,'Mozilla/4.5 (compatible; OmniWeb/4.1.1-v424.6; Mac_PowerPC)','Mac OS','OmniWeb 4.1'),(1261,'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.3a) Gecko/20021207 Phoenix/0.5','Linux','Mozilla 1.3'),(1262,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; MSN 6.1; MSNbMSFT; MSNmnl-nl; MSNc00; v5m)','Windows 98','Internet Explorer 6.0'),(1263,'Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US; rv:1.2.1) Gecko/20021130','Windows XP','Mozilla 1.2'),(1264,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.0; Q312461; ONDOWN3.2; .NET CLR 1.0.3705)','Windows 2000','Internet Explorer 6.0'),(1265,'Mozilla/4.0 (compatible; MSIE 5.5; Windows NT 4.0; YComp 5.0.2.6)','Windows NT 4.0','Internet Explorer 6.0'),(1266,'HTTP agent',NULL,NULL),(1267,'FAST-WebCrawler/3.7 (atw-crawler at fast dot no; http://fast.no/support/crawler.asp)',NULL,NULL),(1268,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; CWIE55NL60)','Windows XP','Internet Explorer 6.0'),(1269,'Mozilla/4.0 (compatible; MSIE 5.16; Mac_PowerPC)','Mac OS','Internet Explorer 5.0'),(1270,'Opera/7.0 (Windows XP; U)  [en]','Windows XP','Opera 7.0'),(1271,'libwww-perl/5.65',NULL,NULL),(1272,'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.0.0) Gecko/20020623 Debian/1.0.0-0.woody.1','Linux','Mozilla 1.0'),(1273,'Mozilla/5.0 (Windows; U; Win98; en-US; rv:1.0.1) Gecko/20020823 Netscape/7.0','Windows 98','Netscape 7.0'),(1274,'Mozilla/5.0 (X11; U; Linux i686; fr-FR; rv:1.0.1) Gecko/20020823 Netscape/7.0','Linux','Netscape 7.0'),(1275,'Mozilla/5.0 (Windows; U; Windows NT 5.1; fr-FR; rv:1.0.1) Gecko/20020823 Netscape/7.0','Windows XP','Netscape 7.0'),(1276,'./a',NULL,NULL),(1277,'Mozilla/4.0 (compatible; MSIE 5.0; Linux 2.4.18-18.8.0 i686) Opera 6.1  [en]','Linux','Opera 6.1'),(1278,'Opera/6.03 (Linux 2.4.20 i686; U)  [en]','Linux','Opera 6.0'),(1279,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; Win 9x 4.90; Hotbar 4.1.8.0)','Windows 98','Internet Explorer 6.0'),(1280,'Windows-Media-Player/9.00.00.2926',NULL,'Windows Media Player 9'),(1281,'Mozilla/4.0 (Windows 2000 5.0) Java/1.4.1_01','Windows 2000','Java 1.4'),(1282,'Zeus 2.6',NULL,NULL),(1283,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; MSN 6.1; MSNbMSFT; MSNmnl-nl; MSNc0z; MSNc00)','Windows XP','Internet Explorer 6.0'),(1284,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; {Vizzavi052001})','Windows 98','Internet Explorer 6.0'),(1285,'Mozilla/4.0 (compatible; MSIE 6.0; MSIE 5.5; Windows 2000) Opera 7.0  [en]','Windows 2000','Opera 7.0'),(1286,'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.1) Gecko/20021023','Linux','Mozilla 1.1'),(1287,'Mozilla/5.0 (Windows; U; Windows NT 5.0; en-US; rv:1.2.1) Gecko/20021130','Windows 2000','Mozilla 1.2'),(1288,'DSNS 1.0',NULL,NULL),(1289,'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.1) Gecko/20020930','Linux','Mozilla 1.1'),(1290,'Opera/6.01 (Windows 98; U)  [en]','Windows 98','Opera 6.0'),(1291,'Links (2.0pre4; Linux 2.4.19-grsec i686; 83x45)','Linux','Links 2.0'),(1292,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; Win 9x 4.90; YComp 5.0.2.6)','Windows 98','Internet Explorer 6.0'),(1293,'WinampMPEG/2.8',NULL,'Winamp 2.8'),(1294,'Mozilla/5.0 (Windows; U; Win98; en-US; rv:1.3a) Gecko/20021212','Windows 98','Mozilla 1.3'),(1295,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; iOpus-I-M)','Windows XP','Internet Explorer 6.0'),(1296,'Mozilla/4.0 (compatible; MSIE 5.01; Windows NT 5.0; Deltamech)','Windows 2000','Internet Explorer 5.0'),(1297,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; Win 9x 4.90; 1&1 Internet AG)','Windows 98','Internet Explorer 6.0'),(1298,'PHP 4.2.2',NULL,NULL),(1299,'PHP 4.2.1',NULL,NULL),(1300,'Java1.3.1_04',NULL,'Java 1.3'),(1301,'Mozilla/4.7 [en]C-CCK-MCD DOJ3jx7bf  (Win95; U)','Windows 95','Netscape 4.7'),(1302,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; FREE)','Windows 98','Internet Explorer 6.0'),(1303,'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; MSOCD; AtHome020)','Windows 98','Internet Explorer 5.5'),(1304,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; Planet Internet PIM 2; Hotbar 4.1.8.0)','Windows XP','Internet Explorer 6.0'),(1305,'Mozilla/5.0 (Windows; U; Windows NT 5.1; it-IT; rv:1.2.1) Gecko/20021130','Windows XP','Mozilla 1.2'),(1306,'Mozilla/5.0 (Windows; U; Windows NT 5.0; en-US; rv:1.3a) Gecko/20021212','Windows 2000','Mozilla 1.3'),(1307,'Mozilla/4.0 (Windows XP 5.1) Java/1.4.1','Windows XP','Java 1.4'),(1308,'Mozilla/4.0 (compatible; MSIE 5.01; Windows 98; T-Online Internatinal AG)','Windows 98','Internet Explorer 5.0'),(1309,'Mozilla/4.0 (Windows 2000 5.0)','Windows 2000',NULL),(1310,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.0; AT&T CSM6.0)','Windows 2000','Internet Explorer 6.0'),(1311,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; MyIE2; .NET CLR 1.0.3705)','Windows XP','Inernet Explorer'),(1312,'Mozilla/4.0 (compatible; MSIE 5.0; Linux) Opera 6.03  [en]','Linux','Opera 6.0'),(1313,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; Win 9x 4.90; FREE)','Windows 98','Internet Explorer 6.0'),(1314,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.0; Q312461; TUCOWS; MyIE2; .NET CLR 1.0.3705)','Windows 2000','Internet Explorer 6.0'),(1315,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; G.A.S.S. 2.0)','Windows XP','Internet Explorer 6.0'),(1316,'Mozilla/5.0 (compatible; Konqueror/3; FreeBSD)','FreeBSD','Konqueror 3.0'),(1317,'Mozilla/5.0 (compatible; Konqueror/3.1; Linux)','Linux','Konqueror 3.1'),(1318,'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.2.1) Gecko/20021218','Linux','Mozilla 1.2'),(1319,'Mozilla/5.0 (X11; U; FreeBSD i386; en-US; rv:1.0.1) Gecko/20030124','FreeBSD','Mozilla 1.0'),(1320,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; MSN 6.1; MSNbMSFT; MSNmnl-nl; MSNc00; v5m)','Windows XP','Internet Explorer 6.0'),(1321,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; .NET CLR 1.0.2914)','Windows XP','Internet Explorer 6.0'),(1322,'Mozilla/2.0 (compatible; MSIE 3.02; Windows CE; PPC; 240x320)','Windows CE','Internet Explorer 3.0'),(1323,'Mozilla/5.0 (X11; U; FreeBSD i386; en-US; rv:1.2.1) Gecko/20030121','FreeBSD','Mozilla 1.2'),(1324,'Links (0.98; Linux 2.4.17 i686; 128x48)','Linux','Links 0.9'),(1325,'Scooter/3.3',NULL,NULL),(1326,'TrafficAnalasysAgent/1.0 (http://trafficanalysis.clickztracker.com/ for more info.)',NULL,NULL),(1327,'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:0.9.9) Gecko/20020408','Linux','Mozilla 0.9'),(1328,'Links (2.1pre7; Linux 2.4.19 i686; x)','Linux','Links 2.1'),(1329,'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.2.1) Gecko/20030111','Linux','Mozilla 1.2'),(1330,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; Multikabel)','Windows 98','Internet Explorer 6.0'),(1331,'Mozilla/5.0 (Macintosh; U; PPC; en-US; rv:1.0.2) Gecko/20021120 Netscape/7.01','Mac OS','Netscape 7.0'),(1332,'Mozilla/5.0 (X11; U; Linux i686; ru-RU; rv:1.2.1) Gecko/20021224','Linux','Mozilla 1.2'),(1333,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; DigExt; WMM)',NULL,NULL),(1334,'Windows-Media-Player/7.00.00.1440',NULL,'Windows Media Player 7'),(1335,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; PABO Haarlem HOME)','Windwows 98','Internet Explorer 6.0'),(1336,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; DigExt; YComp 5.0.0.0; Hotbar 3.0)','Windows XP','Internet Explorer 6.0'),(1337,'Mozilla/4.0 (compatible; MSIE 5.01; Windows NT 5.0; Creative)','Windows 2000','Internet Explorer 5.0'),(1338,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; Hotbar 4.2.6.0)',NULL,NULL),(1339,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; DigExt; Hotbar 4.1.8.0)','Windows 98','Internet Explorer 5.0'),(1340,'Links (0.96; Linux 2.4.7-10 i586)','Linux','Links 0.9'),(1341,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; AtHome033)','Windows XP','Internet Explorer 6.0'),(1342,'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; Hotbar 4.2.4.0)','Windows XP','Internet Explorer 6.0'),(1343,'Mozilla/4.0 (compatible; MSIE 6.0; Windows 98; Planet Internet PIM 2)','Windows 98','Internet Explorer 6.0'),(1344,'Mozilla/5.0 Galeon/1.2.6 (X11; Linux i686; U;) Gecko/20020913','Linux','Galeon 1.2'),(1345,'Mozilla/4.0 (compatible; MSIE 5.0; Windows 95; DigExt; btclick.com Build BTCFSOHOQ2)',NULL,NULL);

